import dataclasses
import os
import typing
from collections import defaultdict

import numpy as np
import pytest

from PartSeg._roi_analysis.partseg_settings import PartSettings
from PartSeg._roi_mask.main_window import ChosenComponents
from PartSeg._roi_mask.stack_settings import StackSettings, get_mask
from PartSeg.common_backend.base_settings import BaseSettings, SwapTimeStackException, TimeAndStackException
from PartSegCore.algorithm_describe_base import ROIExtractionProfile
from PartSegCore.analysis import AnalysisAlgorithmSelection
from PartSegCore.analysis.io_utils import MaskInfo, create_history_element_from_project
from PartSegCore.analysis.load_functions import LoadProject
from PartSegCore.analysis.save_functions import SaveProject
from PartSegCore.io_utils import PointsInfo
from PartSegCore.mask.history_utils import create_history_element_from_segmentation_tuple
from PartSegCore.mask.io_functions import LoadStackImage
from PartSegCore.project_info import HistoryProblem, calculate_mask_from_project
from PartSegCore.roi_info import ROIInfo
from PartSegCore.segmentation.algorithm_base import ROIExtractionResult
from PartSegImage import Image


@pytest.fixture()
def stack_settings(qtbot, tmp_path):
    settings = StackSettings(tmp_path)
    chose = ChosenComponents()
    qtbot.addWidget(chose)
    settings.chosen_components_widget = chose
    return settings


class TestStackSettings:
    def test_transform_state_simple(self, stack_image):
        roi = np.zeros(stack_image.image.get_channel(0).shape, dtype=np.uint8)
        roi[0, 1, 2:-2, 2:-2] = 1
        roi[0, 2, 2:-2, 2:-2] = 2
        roi[0, 3, 2:-2, 2:-2] = 3
        roi_info = ROIInfo(roi).fit_to_image(stack_image.image)
        roi_extraction_parameters = defaultdict(
            lambda: ROIExtractionProfile(name="aa", algorithm="aa", values={1: "aa"})
        )
        new_state = StackSettings.transform_state(
            state=stack_image,
            new_roi_info=roi_info,
            new_roi_extraction_parameters=roi_extraction_parameters,
            list_of_components=[2],
        )
        assert len(new_state.roi_extraction_parameters) == 3
        assert new_state.selected_components == [2]

    def test_transform_state(self, stack_segmentation1):
        roi_extraction_parameters = defaultdict(
            lambda: ROIExtractionProfile(name="aa", algorithm="aa", values={1: "aa"})
        )
        new_state = StackSettings.transform_state(
            state=stack_segmentation1,
            new_roi_info=stack_segmentation1.roi_info,
            new_roi_extraction_parameters=roi_extraction_parameters,
            list_of_components=[4],
        )
        assert new_state.selected_components == [1, 2, 4]
        assert new_state.roi_extraction_parameters[1] == stack_segmentation1.roi_extraction_parameters[1]
        assert new_state.roi_extraction_parameters[2] == stack_segmentation1.roi_extraction_parameters[3]
        assert new_state.roi_extraction_parameters[3] == ROIExtractionProfile(
            name="aa", algorithm="aa", values={1: "aa"}
        )
        assert np.all((new_state.roi_info.roi == 2) == (stack_segmentation1.roi_info.roi == 3))

    def test_add_project(self, stack_settings, stack_segmentation1, data_test_dir):
        stack_settings.set_project_info(stack_segmentation1)
        project1_res = stack_settings.get_project_info()
        assert isinstance(project1_res.selected_components, typing.Iterable)
        assert len(project1_res.selected_components) == 2
        assert project1_res.file_path == "test_path"
        project2 = LoadStackImage.load([os.path.join(data_test_dir, "test_lsm.tif")])
        stack_settings.set_project_info(project2)
        project2_res = stack_settings.get_project_info()
        assert project2_res.roi_info.roi is None
        assert isinstance(project2_res.selected_components, typing.Iterable)
        assert len(project2_res.selected_components) == 0
        assert project2_res.file_path == os.path.join(data_test_dir, "test_lsm.tif")

    def test_set_project(self, stack_settings, stack_segmentation1, stack_segmentation2):
        stack_settings.set_project_info(stack_segmentation1)
        segmentation1 = stack_settings.roi
        stack_settings.set_project_info(stack_segmentation2)
        segmentation2 = stack_settings.roi
        assert np.any(segmentation1 != segmentation2)

    def test_set_project_with_components(self, stack_settings, stack_segmentation1):
        seg1 = dataclasses.replace(stack_segmentation1, selected_components=[1, 2])
        seg2 = dataclasses.replace(stack_segmentation1, selected_components=[3])
        stack_settings.set_project_info(seg1)
        assert stack_settings.chosen_components() == [1, 2]
        stack_settings.set_project_info(seg2)
        assert stack_settings.chosen_components() == [3]
        stack_settings.keep_chosen_components = True
        stack_settings.set_project_info(seg1)
        assert stack_settings.chosen_components() == [1, 2, 3]

    def test_set_project_with_history(self, stack_settings, stack_segmentation1, mask_property):
        seg2 = dataclasses.replace(
            stack_segmentation1,
            history=[create_history_element_from_segmentation_tuple(stack_segmentation1, mask_property)],
            selected_components=[1],
        )
        seg3 = dataclasses.replace(
            stack_segmentation1,
            history=[create_history_element_from_segmentation_tuple(stack_segmentation1, mask_property)],
            selected_components=[2],
        )
        stack_settings.set_project_info(seg2)
        assert stack_settings.chosen_components() == [1]
        stack_settings.set_project_info(seg3)
        assert stack_settings.chosen_components() == [2]
        stack_settings.keep_chosen_components = True
        stack_settings.set_project_info(seg2)
        assert stack_settings.chosen_components() == [1, 2]

    def test_set_project_with_history_components_fail(self, stack_settings, stack_segmentation1, mask_property):
        seg2 = dataclasses.replace(
            stack_segmentation1,
            history=[
                create_history_element_from_segmentation_tuple(
                    dataclasses.replace(stack_segmentation1, selected_components=[1]), mask_property
                )
            ],
            selected_components=[1],
        )
        seg3 = dataclasses.replace(
            stack_segmentation1,
            history=[
                create_history_element_from_segmentation_tuple(
                    dataclasses.replace(stack_segmentation1, selected_components=[1, 2]), mask_property
                )
            ],
            selected_components=[2],
        )
        stack_settings.set_project_info(seg2)
        assert stack_settings.chosen_components() == [1]
        stack_settings.set_project_info(seg3)
        assert stack_settings.chosen_components() == [2]
        stack_settings.keep_chosen_components = True
        with pytest.raises(HistoryProblem):
            stack_settings.set_project_info(seg2)

    def test_set_project_with_history_parameters_fail(
        self, stack_settings, stack_segmentation1, stack_segmentation2, mask_property
    ):
        seg2 = dataclasses.replace(
            stack_segmentation1,
            history=[create_history_element_from_segmentation_tuple(stack_segmentation1, mask_property)],
            selected_components=[1],
        )
        seg3 = dataclasses.replace(
            stack_segmentation1,
            history=[create_history_element_from_segmentation_tuple(stack_segmentation2, mask_property)],
            selected_components=[2],
        )
        stack_settings.set_project_info(seg2)
        assert stack_settings.chosen_components() == [1]
        stack_settings.set_project_info(seg3)
        assert stack_settings.chosen_components() == [2]
        stack_settings.keep_chosen_components = True
        with pytest.raises(HistoryProblem):
            stack_settings.set_project_info(seg2)

    def test_set_project_with_history_length_fail(self, stack_settings, stack_segmentation1, mask_property):
        seg2 = dataclasses.replace(
            stack_segmentation1,
            history=[
                create_history_element_from_segmentation_tuple(stack_segmentation1, mask_property),
                create_history_element_from_segmentation_tuple(stack_segmentation1, mask_property),
            ],
            selected_components=[1],
        )
        seg3 = dataclasses.replace(
            stack_segmentation1,
            history=[create_history_element_from_segmentation_tuple(stack_segmentation1, mask_property)],
            selected_components=[2],
        )
        stack_settings.set_project_info(seg2)
        assert stack_settings.chosen_components() == [1]
        stack_settings.set_project_info(seg3)
        assert stack_settings.chosen_components() == [2]
        stack_settings.keep_chosen_components = True
        with pytest.raises(HistoryProblem):
            stack_settings.set_project_info(seg2)

    def test_set_segmentation_result(self, stack_settings, stack_segmentation1, stack_image):
        stack_settings.set_project_info(stack_image)
        seg = ROIExtractionResult(
            roi=stack_segmentation1.roi_info.roi,
            parameters=ROIExtractionProfile(name="test", algorithm="test2", values={}),
        )
        stack_settings.set_segmentation_result(seg)
        assert stack_settings.last_executed_algorithm == "test2"
        assert np.array_equal(stack_settings.roi, stack_segmentation1.roi_info.roi)

    def test_selected_components(self, stack_settings, stack_segmentation1):
        stack_settings.set_project_info(stack_segmentation1)
        assert stack_settings.chosen_components() == [1, 3]
        assert stack_settings.component_is_chosen(1)
        assert not stack_settings.component_is_chosen(2)
        assert np.array_equal(stack_settings.components_mask(), [0, 1, 0, 1, 0])
        stack_settings.chosen_components_widget.un_check_all()
        assert stack_settings.chosen_components() == []
        stack_settings.chosen_components_widget.check_all()
        assert stack_settings.chosen_components() == list(range(1, stack_segmentation1.roi_info.roi.max() + 1))


class TestBaseSettings:
    def test_empty_history(self, tmp_path):
        settings = BaseSettings(tmp_path)
        assert settings.history_size() == 0
        assert settings.history_redo_size() == 0

        with pytest.raises(IndexError):
            settings.history_current_element()

        with pytest.raises(IndexError):
            settings.history_next_element()

        settings.history_pop()
        assert settings.history_index == -1

    def test_modifying_history(self, tmp_path):
        settings = BaseSettings(tmp_path)
        for i in range(10):
            settings.add_history_element(i)

        assert settings.history_size() == 10
        assert settings.history_redo_size() == 0

        with pytest.raises(IndexError):
            settings.history_next_element()

        for i in range(9, 5, -1):
            assert settings.history_current_element() == i
            assert settings.history_pop() == i

        assert settings.history_current_element() == 5
        assert settings.history_size() == 6
        assert settings.history_redo_size() == 4
        settings.add_history_element(7)
        assert settings.history_size() == 7
        assert settings.history_redo_size() == 0

    def test_clean_redo(self, tmp_path):
        settings = BaseSettings(tmp_path)
        for i in range(10):
            settings.add_history_element(i)
        for _ in range(5):
            settings.history_pop()
        settings.history_current_element()
        settings.history_redo_clean()
        settings.history_current_element()

    def test_add_point(self, tmp_path, qtbot):
        settings = BaseSettings(tmp_path)
        with qtbot.waitSignal(settings.points_changed):
            settings.points = [1, 2, 3]

    def test_set_roi(self, tmp_path, qtbot):
        settings = BaseSettings(tmp_path)
        roi = np.zeros((10, 10), dtype=np.uint8)
        settings.image = Image(roi, (1, 1), axes_order="XY")
        roi[1:5, 1:5] = 1
        roi[5:-1, 5:-1] = 3
        with qtbot.waitSignal(settings.roi_changed):
            settings.roi = roi
        assert len(settings.roi_info.bound_info) == 2
        assert set(settings.roi_info.bound_info) == {1, 3}
        assert settings.roi_info.alternative == {}
        assert settings.roi_info.annotations == {}

        with qtbot.waitSignal(settings.roi_clean):
            settings.roi = None
        assert settings.roi is None

        settings.image = None
        assert settings.image is not None

    def test_channels(self, tmp_path, qtbot):
        settings = BaseSettings(tmp_path)
        assert not settings.has_channels
        assert settings.channels == 0
        settings.image = Image(np.zeros((10, 10, 2), dtype=np.uint8), (1, 1), axes_order="XYC")
        assert settings.has_channels
        assert settings.channels == 2
        settings.image = Image(np.zeros((10, 10, 1), dtype=np.uint8), (1, 1), axes_order="XYC")
        assert not settings.has_channels
        assert settings.channels == 1

    def test_shape(self, tmp_path):
        settings = BaseSettings(tmp_path)
        assert settings.image_shape == ()
        settings.image = Image(np.zeros((10, 10, 2), dtype=np.uint8), (1, 1), axes_order="XYC")
        assert settings.image_shape == (1, 1, 10, 10)

    def test_verify_image(self):
        assert BaseSettings.verify_image(Image(np.zeros((10, 10, 2), dtype=np.uint8), (1, 1), axes_order="XYC"))
        with pytest.raises(SwapTimeStackException):
            BaseSettings.verify_image(
                Image(np.zeros((2, 10, 10), dtype=np.uint8), (1, 1, 1), axes_order="TXY"), silent=False
            )
        im = BaseSettings.verify_image(Image(np.zeros((2, 10, 10), dtype=np.uint8), (1, 1, 1), axes_order="TXY"))
        assert not im.is_time
        assert im.times == 1
        assert im.is_stack
        assert im.layers == 2
        with pytest.raises(TimeAndStackException):
            BaseSettings.verify_image(Image(np.zeros((2, 2, 10, 10), dtype=np.uint8), (1, 1, 1), axes_order="TZXY"))

    def test_algorithm_redirect(self, tmp_path):
        settings = BaseSettings(tmp_path)
        settings.set_algorithm("algorithms.aa", 2)
        with pytest.warns(FutureWarning, match="Use `set_algorithm_state` instead"):
            assert settings.get("algorithms.aa") == 2
        with pytest.warns(FutureWarning, match="Use `set_algorithm_state` instead"):
            settings.set("algorithms.aa", 3)
        assert settings.get_algorithm("algorithms.aa") == 3


class TestPartSettings:
    def test_set_mask_info(self, qtbot, tmp_path, image):
        settings = PartSettings(tmp_path)
        settings.image = image
        mask_info = MaskInfo("", (image.get_channel(0) > 2).astype(np.uint8))
        with qtbot.wait_signal(settings.mask_changed):
            settings.set_project_info(mask_info)

    def test_project_info_set(self, qtbot, analysis_segmentation, analysis_segmentation2, tmp_path, image):
        settings = PartSettings(tmp_path)
        settings.image = image
        settings.set_project_info(analysis_segmentation)
        assert settings.mask is None
        settings.set_project_info(analysis_segmentation2)
        assert settings.mask is not None
        settings.set_project_info(analysis_segmentation)
        assert settings.mask is None
        analysis_segmentation3 = dataclasses.replace(analysis_segmentation2, roi_info=ROIInfo(None))
        settings.set_project_info(analysis_segmentation3)
        assert settings.mask is not None
        assert settings.roi is None

    def test_get_project_info(self, qtbot, tmp_path, image):
        settings = PartSettings(tmp_path)
        settings.last_executed_algorithm = "aa"
        settings.set_algorithm("algorithms.aa", 1)
        settings.image = image
        pt = settings.get_project_info()
        assert pt.algorithm_parameters["algorithm_name"] == "aa"
        assert pt.algorithm_parameters["values"] == 1

    def test_pipeline_saving(self, qtbot, tmp_path, image, algorithm_parameters, mask_property):
        settings = PartSettings(tmp_path)
        settings.image = image
        settings.last_executed_algorithm = algorithm_parameters["algorithm_name"]
        algorithm = AnalysisAlgorithmSelection[algorithm_parameters["algorithm_name"]]()
        algorithm.set_image(settings.image)
        algorithm.set_parameters(**algorithm_parameters["values"])
        result = algorithm.calculation_run(lambda x, y: None)
        settings.set_segmentation_result(result)
        project_info = settings.get_project_info()
        mask = calculate_mask_from_project(mask_property, settings.get_project_info())
        settings.add_history_element(
            create_history_element_from_project(
                project_info,
                mask_property,
            )
        )
        settings.mask = mask
        calculate_mask_from_project(mask_property, settings.get_project_info())
        algorithm_parameters["values"]["channel"] = 1
        algorithm.set_parameters(**algorithm_parameters["values"])
        algorithm.set_mask(settings.mask)
        result2 = algorithm.calculation_run(lambda x, y: None)
        assert np.max(result2.roi) == 2
        settings.set_segmentation_result(result2)
        project_info = settings.get_project_info()
        SaveProject.save(tmp_path / "project.tgz", project_info)
        assert os.path.exists(tmp_path / "project.tgz")
        loaded = LoadProject.load([tmp_path / "project.tgz"])
        assert np.all(loaded.roi_info.roi == result2.roi)
        assert len(loaded.history) == 1


@pytest.mark.parametrize("Settings", [PartSettings, StackSettings])
def test_set_project_info(Settings, qtbot, tmp_path, image):
    settings = Settings(tmp_path)
    settings.points = [1, 2, 3]
    assert settings.points == [1, 2, 3]
    settings.image = image
    assert settings.points is None
    settings.set_project_info(PointsInfo("a", np.arange(3)))
    assert np.all(settings.points == [0, 1, 2])


def test_get_mask(stack_segmentation1):
    res1 = get_mask(stack_segmentation1.roi_info.roi, mask=None, selected=[1, 3])
    assert isinstance(res1, np.ndarray)
    assert set(np.unique(res1)) == {0, 1}
    assert set(stack_segmentation1.roi_info.roi[res1 > 0]) == {0, 2, 4}
    mask = np.ones(stack_segmentation1.roi_info.roi.shape, dtype=np.uint8)
    res2 = get_mask(stack_segmentation1.roi_info.roi, mask=mask, selected=[1, 3])
    assert set(np.unique(res2)) == {0, 1}
    assert np.array_equal(res1, res2)
    res3 = get_mask(stack_segmentation1.roi_info.roi, mask=res1, selected=[1, 2])
    assert set(stack_segmentation1.roi_info.roi[res3 > 0]) == {0, 4}
    assert set(np.unique(res3)) == {0, 1}
    res4 = get_mask(stack_segmentation1.roi_info.roi, mask=None, selected=[1])
    assert np.array_equal(res4, stack_segmentation1.roi_info.roi != 1)

    assert get_mask(stack_segmentation1.roi_info.roi, mask=None, selected=[]) is None
    assert get_mask(None, mask=None, selected=[1, 2]) is None
    assert np.array_equal(get_mask(None, mask=res4, selected=[1, 2]), res4)
    assert np.array_equal(get_mask(stack_segmentation1.roi_info.roi, mask=res4, selected=[]), res4)



# pylint: disable=no-self-use

import dataclasses
import json
import os.path
import re
import tarfile
from copy import deepcopy
from enum import Enum
from glob import glob
from io import BytesIO
from pathlib import Path
from typing import Type

import h5py
import numpy as np
import pandas as pd
import pytest
import tifffile

from PartSegCore import UNIT_SCALE, Units
from PartSegCore.algorithm_describe_base import ROIExtractionProfile
from PartSegCore.analysis import ProjectTuple
from PartSegCore.analysis.calculation_plan import CalculationPlan, MaskSuffix, MeasurementCalculate
from PartSegCore.analysis.load_functions import LoadImageForBatch, LoadProject
from PartSegCore.analysis.measurement_base import Leaf, MeasurementEntry
from PartSegCore.analysis.measurement_calculation import MEASUREMENT_DICT, MeasurementProfile
from PartSegCore.analysis.save_functions import SaveAsNumpy, SaveAsTiff, SaveCmap, SaveProject, SaveXYZ
from PartSegCore.io_utils import (
    LoadBase,
    LoadPlanExcel,
    LoadPlanJson,
    LoadPoints,
    SaveBase,
    SaveMaskAsTiff,
    SaveROIAsNumpy,
    SaveScreenshot,
    find_problematic_entries,
    find_problematic_leafs,
    load_metadata_base,
    load_metadata_part,
    open_tar_file,
)
from PartSegCore.json_hooks import PartSegEncoder, partseg_object_hook
from PartSegCore.mask.history_utils import create_history_element_from_segmentation_tuple
from PartSegCore.mask.io_functions import (
    LoadROI,
    LoadROIFromTIFF,
    LoadROIImage,
    LoadROIParameters,
    LoadStackImage,
    LoadStackImageWithMask,
    MaskProjectTuple,
    SaveComponents,
    SaveParametersJSON,
    SaveROI,
    save_components,
)
from PartSegCore.mask_create import MaskProperty
from PartSegCore.project_info import HistoryElement, ProjectInfoBase
from PartSegCore.roi_info import ROIInfo
from PartSegCore.segmentation.algorithm_base import AdditionalLayerDescription
from PartSegCore.segmentation.noise_filtering import DimensionType
from PartSegCore.segmentation.segmentation_algorithm import ThresholdAlgorithm
from PartSegCore.segmentation.threshold import RangeThresholdSelection
from PartSegCore.utils import ProfileDict, check_loaded_dict
from PartSegImage import Image


@pytest.fixture(scope="module")
def analysis_project() -> ProjectTuple:
    data = np.zeros((1, 1, 50, 100, 100), dtype=np.uint16)
    data[0, 0, 10:40, 10:40, 10:90] = 50
    data[0, 0, 10:40, 50:90, 10:90] = 50
    data[0, 0, 15:35, 15:35, 15:85] = 70
    data[0, 0, 15:35, 55:85, 15:85] = 60
    data[0, 0, 10:40, 40:50, 10:90] = 40
    image = Image(
        data,
        (10 / UNIT_SCALE[Units.nm.value], 5 / UNIT_SCALE[Units.nm.value], 5 / UNIT_SCALE[Units.nm.value]),
        "",
        axes_order="CTZYX",
    )
    mask = data[0, 0] > 0
    roi = np.zeros(data.shape, dtype=np.uint8)
    roi[data == 70] = 1
    roi[data == 60] = 2
    algorithm_parameters = {
        "algorithm_name": "Lower Threshold",
        "values": {
            "threshold": 60,
            "channel": 0,
            "noise_filtering": {"name": "None", "values": {}},
            "minimum_size": 10,
            "side_connection": False,
        },
    }
    roi_info = ROIInfo(roi.squeeze()).fit_to_image(image)
    return ProjectTuple(
        file_path="test_data.tiff",
        image=image,
        roi_info=roi_info,
        additional_layers={"denoised image": AdditionalLayerDescription(data=roi, layer_type="layer")},
        mask=mask,
        algorithm_parameters=algorithm_parameters,
    )


@pytest.fixture(scope="module")
def analysis_project_reversed() -> ProjectTuple:
    data = np.zeros((1, 1, 50, 100, 100), dtype=np.uint16)
    data[0, 0, 10:40, 10:40, 10:90] = 50
    data[0, 0, 10:40, 50:90, 10:90] = 50
    data[0, 0, 15:35, 15:35, 15:85] = 70
    data[0, 0, 15:35, 55:85, 15:85] = 60
    data[0, 0, 10:40, 40:50, 10:90] = 40
    mask = data[0, 0] > 0
    roi = np.zeros(data.shape, dtype=np.uint8)
    roi[data == 70] = 1
    roi[data == 60] = 2

    data = 100 - data
    image = Image(
        data,
        (10 / UNIT_SCALE[Units.nm.value], 5 / UNIT_SCALE[Units.nm.value], 5 / UNIT_SCALE[Units.nm.value]),
        "",
        axes_order="CTZYX",
    )
    roi_info = ROIInfo(roi.squeeze()).fit_to_image(image)
    return ProjectTuple("test_data.tiff", image, roi_info=roi_info, mask=mask)


@pytest.fixture()
def mask_prop():
    return MaskProperty.simple_mask()


class SampleEnumClass(Enum):
    test0 = 0
    test1 = 1
    test2 = 2


class TestHistoryElement:
    def test_create(self, mask_prop):
        roi_info = ROIInfo(np.zeros((10, 10), dtype=np.uint8))
        elem = HistoryElement.create(roi_info, None, {}, mask_prop)
        assert elem.mask_property == mask_prop
        assert elem.roi_extraction_parameters == {}
        param = {"a": 1, "b": 2}
        elem2 = HistoryElement.create(roi_info, None, param, mask_prop)
        assert elem2.roi_extraction_parameters == param
        roi_info2, mask = elem2.get_roi_info_and_mask()
        assert np.all(roi_info2.roi == 0)
        assert mask is None

    def test_mask(self, mask_prop):
        mask = np.zeros((10, 10), dtype=np.uint8)
        mask[1:-1, 1:-1] = 2
        mask2 = np.copy(mask)
        mask2[2:-2, 2:-2] = 4
        roi_info = ROIInfo(mask2)
        elem = HistoryElement.create(roi_info, mask, {}, mask_prop)
        roi_info2, mask2 = elem.get_roi_info_and_mask()
        assert np.all(mask == mask2)
        assert np.all(roi_info.roi == roi_info2.roi)

    def test_additional(self, mask_prop):
        data = np.zeros((10, 10), dtype=np.uint8)
        data[1:-1, 1:-1] = 2
        alternative = {}
        for i in range(4):
            alternative_array = np.copy(data)
            alternative_array[alternative_array > 0] = i + 2
            alternative[f"add{i}"] = alternative_array
        roi_info = ROIInfo(data, alternative=alternative)
        elem = HistoryElement.create(roi_info, None, {}, mask_prop)
        roi_info2, mask2 = elem.get_roi_info_and_mask()
        assert np.all(roi_info2.roi == roi_info.roi)
        assert mask2 is None
        assert len(roi_info.alternative) == 4
        assert set(roi_info.alternative) == {f"add{i}" for i in range(4)}
        for i in range(4):
            arr = roi_info.alternative[f"add{i}"]
            assert np.all(arr[arr > 0] == i + 2)

    def test_annotations(self, mask_prop):
        data = np.zeros((10, 10), dtype=np.uint8)
        data[1:5, 1:5] = 1
        data[5:-1, 1:5] = 2
        data[1:5, 5:-1] = 3
        data[5:-1, 5:-1] = 4
        annotations = {1: "a", 2: "b", 3: "c", 4: "d"}
        roi_info = ROIInfo(data, annotations=annotations)
        elem = HistoryElement.create(roi_info, None, {}, mask_prop)
        roi_info2, mask2 = elem.get_roi_info_and_mask()
        assert mask2 is None
        assert np.all(roi_info2.roi == roi_info.roi)


class TestSaveHistory:
    def test_save_roi_info_project_tuple(self, analysis_segmentation2, tmp_path):
        self.perform_roi_info_test(analysis_segmentation2, tmp_path, SaveProject, LoadProject)

    def test_save_roi_info_mask_project(self, stack_segmentation2, tmp_path):
        self.perform_roi_info_test(stack_segmentation2, tmp_path, SaveROI, LoadROI)

    def perform_roi_info_test(self, project, save_path, save_method: Type[SaveBase], load_method: Type[LoadBase]):
        assert save_method.get_short_name().lower() == save_method.get_short_name()
        assert save_method.get_short_name().isalpha()
        assert load_method.get_short_name().lower() == load_method.get_short_name()
        assert load_method.get_short_name().isalpha()
        alt1 = np.copy(project.roi_info.roi)
        alt1[alt1 > 0] += 3
        roi_info = ROIInfo(
            roi=project.roi_info.roi, annotations={i: f"a{i}" for i in range(1, 5)}, alternative={"test": alt1}
        )
        proj = dataclasses.replace(project, roi_info=roi_info)
        save_method.save(save_path / "data.tgz", proj, SaveROI.get_default_values())
        proj2 = load_method.load([save_path / "data.tgz"])
        assert np.all(proj2.roi_info.roi == project.roi_info.roi)
        assert set(proj2.roi_info.annotations) == {1, 2, 3, 4}
        assert proj2.roi_info.annotations == {i: f"a{i}" for i in range(1, 5)}
        assert "test" in proj2.roi_info.alternative
        assert np.all(proj2.roi_info.alternative["test"] == alt1)

    def test_save_roi_info_history_project_tuple(self, analysis_segmentation2, mask_property, tmp_path):
        self.perform_roi_info_history_test(analysis_segmentation2, tmp_path, mask_property, SaveProject, LoadProject)

    def test_save_roi_info_history_mask_project(self, stack_segmentation2, mask_property, tmp_path):
        self.perform_roi_info_history_test(stack_segmentation2, tmp_path, mask_property, SaveROI, LoadROI)

    def perform_roi_info_history_test(
        self, project, save_path, mask_property, save_method: Type[SaveBase], load_method: Type[LoadBase]
    ):
        alt1 = np.copy(project.roi_info.roi)
        alt1[alt1 > 0] += 3
        roi_info = ROIInfo(
            roi=project.roi_info.roi, annotations={i: f"a{i}" for i in range(1, 5)}, alternative={"test": alt1}
        )
        history = []
        for i in range(3):
            alt2 = np.copy(alt1)
            alt2[alt2 > 0] = i + 5
            roi_info2 = ROIInfo(
                roi=project.roi_info.roi,
                annotations={j: f"a{i}_{j}" for j in range(1, 5)},
                alternative={f"test{i}": alt2},
            )
            history.append(
                HistoryElement.create(
                    roi_info2, alt1, {"algorithm_name": f"task_{i}", "values": {"a": 1}}, mask_property
                )
            )
        proj = dataclasses.replace(project, roi_info=roi_info, history=history)
        save_method.save(save_path / "data.tgz", proj, SaveROI.get_default_values())
        proj2: ProjectInfoBase = load_method.load([save_path / "data.tgz"])
        assert np.all(proj2.roi_info.roi == project.roi_info.roi)
        assert set(proj2.roi_info.annotations) == {1, 2, 3, 4}
        assert proj2.roi_info.annotations == {i: f"a{i}" for i in range(1, 5)}
        assert "test" in proj2.roi_info.alternative
        assert np.all(proj2.roi_info.alternative["test"] == alt1)
        assert len(proj2.history) == 3
        for i in range(3):
            roi_info3, mask2 = proj2.history[i].get_roi_info_and_mask()
            assert np.all(mask2 == alt1)
            assert set(roi_info3.alternative) == {f"test{i}"}
            assert np.all(roi_info3.alternative[f"test{i}"][alt1 > 0] == i + 5)
            assert np.all(roi_info3.alternative[f"test{i}"][alt1 == 0] == 0)
            assert roi_info3.annotations == {j: f"a{i}_{j}" for j in range(1, 5)}
            assert proj2.history[i].roi_extraction_parameters == {"algorithm_name": f"task_{i}", "values": {"a": 1}}


class TestJsonLoad:
    def test_profile_load(self, data_test_dir):
        profile_path = os.path.join(data_test_dir, "segment_profile_test.json")
        # noinspection PyBroadException
        try:
            with open(profile_path) as ff:
                data = json.load(ff, object_hook=partseg_object_hook)
            assert check_loaded_dict(data)
        except Exception:  # pylint: disable=broad-except  # pragma: no cover
            pytest.fail("Fail in loading profile")

    def test_measure_load(self, data_test_dir):
        profile_path = os.path.join(data_test_dir, "measurements_profile_test.json")
        # noinspection PyBroadException
        try:
            with open(profile_path) as ff:
                data = json.load(ff, object_hook=partseg_object_hook)
            assert check_loaded_dict(data)
        except Exception:  # pylint: disable=broad-except  # pragma: no cover
            pytest.fail("Fail in loading profile")

    def test_json_dump(self):
        with pytest.raises(TypeError):
            json.dumps(DimensionType.Layer)
        data_string = json.dumps(DimensionType.Layer, cls=PartSegEncoder)
        assert re.search('"__class__":[^,}]+[,}]', data_string) is not None
        assert re.search('"value":[^,}]+[,}]', data_string) is not None

    def test_json_load(self):
        test_json = json.dumps(SampleEnumClass.test0, cls=PartSegEncoder)
        assert isinstance(json.loads(test_json, object_hook=partseg_object_hook), SampleEnumClass)

    def test_modernize_0_9_2_3(self, bundle_test_dir):
        file_path = os.path.join(bundle_test_dir, "segment_profile_0.9.2.3.json")
        assert os.path.exists(file_path)
        data = load_metadata_base(file_path)
        assert hasattr(data["test_0.9.2.3"].values, "noise_filtering")
        assert hasattr(data["test_0.9.2.3"].values.noise_filtering.values, "dimension_type")
        file_path = os.path.join(bundle_test_dir, "calculation_plan_0.9.2.3.json")
        data = load_metadata_base(file_path)

    def test_update_name(self):
        data = load_metadata_base(UPDATE_NAME_JSON)
        mp = data["problematic set"]
        assert isinstance(mp, MeasurementProfile)
        assert isinstance(mp.chosen_fields[0], MeasurementEntry)
        assert isinstance(mp.chosen_fields[0].calculation_tree, Leaf)
        assert mp.chosen_fields[0].calculation_tree.name == "Pixel brightness sum"
        assert mp.chosen_fields[1].calculation_tree.name == "Components number"

    def test_load_workflow(self, bundle_test_dir):
        data = load_metadata_base(os.path.join(bundle_test_dir, "workflow.json"))
        plan = data["workflow"]
        assert isinstance(plan, CalculationPlan)
        mask_step = plan.execution_tree.children[0]
        assert isinstance(mask_step.operation, MaskSuffix)
        segmentation_step = mask_step.children[0]
        assert isinstance(segmentation_step.operation, ROIExtractionProfile)
        measurement_step = segmentation_step.children[0]
        assert isinstance(measurement_step.operation, MeasurementCalculate)
        assert measurement_step.children == []
        measurement_profile = measurement_step.operation.measurement_profile
        assert isinstance(measurement_profile, MeasurementProfile)
        for entry in measurement_profile.chosen_fields:
            assert entry.calculation_tree.name in MEASUREMENT_DICT

    def test_load_workflow_from_text(self, bundle_test_dir):
        with open(os.path.join(bundle_test_dir, "workflow.json")) as ff:
            data_text = ff.read()
        assert isinstance(load_metadata_base(data_text)["workflow"], CalculationPlan)
        with open(os.path.join(bundle_test_dir, "workflow.json")) as ff:
            isinstance(load_metadata_base(ff)["workflow"], CalculationPlan)


class TestSegmentationMask:
    def test_load_seg(self, data_test_dir):
        seg = LoadROI.load([os.path.join(data_test_dir, "test_nucleus_1_1.seg")])
        assert isinstance(seg.image, str)
        assert seg.selected_components == [1, 3]
        assert os.path.exists(os.path.join(data_test_dir, seg.image))
        assert len(seg.roi_extraction_parameters) == 4
        assert os.path.basename(seg.image) == "test_nucleus.tif"

    def test_load_old_seg(self, data_test_dir):
        """
        For PartSeg 0.9.4 and older
        """
        seg = LoadROI.load([os.path.join(data_test_dir, "test_nucleus.seg")])
        assert isinstance(seg.image, str)
        assert seg.selected_components == [1, 3]
        assert os.path.exists(os.path.join(data_test_dir, seg.image))
        assert os.path.basename(seg.image) == "test_nucleus.tif"

    def test_load_old_seg_with_image(self, data_test_dir):
        seg = LoadROIImage.load(
            [os.path.join(data_test_dir, "test_nucleus.seg")], metadata={"default_spacing": (1, 1, 1)}
        )
        assert isinstance(seg.image, Image)
        assert seg.selected_components == [1, 3]
        assert isinstance(seg.roi_info.roi, np.ndarray)
        seg.image.fit_array_to_image(seg.roi_info.roi)
        assert os.path.basename(seg.image.file_path) == "test_nucleus.tif"

    def test_load_seg_with_image(self, data_test_dir):
        seg = LoadROIImage.load(
            [os.path.join(data_test_dir, "test_nucleus_1_1.seg")], metadata={"default_spacing": (1, 1, 1)}
        )
        assert isinstance(seg.image, Image)
        assert seg.selected_components == [1, 3]
        assert isinstance(seg.roi_info.roi, np.ndarray)
        seg.image.fit_array_to_image(seg.roi_info.roi)
        assert os.path.basename(seg.image.file_path) == "test_nucleus.tif"

    def test_save_segmentation(self, tmpdir, data_test_dir):
        seg = LoadROIImage.load(
            [os.path.join(data_test_dir, "test_nucleus_1_1.seg")], metadata={"default_spacing": (1, 1, 1)}
        )
        SaveROI.save(os.path.join(tmpdir, "segmentation.seg"), seg, {"relative_path": False})
        assert os.path.exists(os.path.join(tmpdir, "segmentation.seg"))
        os.makedirs(os.path.join(tmpdir, "seg_save"))
        save_components(
            seg.image,
            seg.selected_components,
            os.path.join(tmpdir, "seg_save"),
            seg.roi_info,
            SaveComponents.get_default_values(),
        )
        assert os.path.isdir(os.path.join(tmpdir, "seg_save"))
        assert len(glob(os.path.join(tmpdir, "seg_save", "*"))) == 4
        seg2 = LoadROI.load([os.path.join(tmpdir, "segmentation.seg")])
        assert seg2 is not None
        save_components(
            seg.image,
            [],
            os.path.join(tmpdir, "seg_save2"),
            seg.roi_info,
            SaveComponents.get_default_values(),
        )
        assert os.path.isdir(os.path.join(tmpdir, "seg_save2"))
        assert len(glob(os.path.join(tmpdir, "seg_save2", "*"))) == 8

    def test_save_segmentation_without_image(self, tmpdir, data_test_dir):
        seg = LoadROIImage.load(
            [os.path.join(data_test_dir, "test_nucleus_1_1.seg")], metadata={"default_spacing": (1, 1, 1)}
        )
        seg_clean = dataclasses.replace(seg, image=None, roi_info=seg.roi_info)
        SaveROI.save(os.path.join(tmpdir, "segmentation.seg"), seg_clean, {"relative_path": False})
        SaveROI.save(
            os.path.join(tmpdir, "segmentation1.seg"),
            seg_clean,
            {"relative_path": False, "spacing": (210 * 10**-6, 70 * 10**-6, 70 * 10**-6)},
        )

    def test_loading_new_segmentation(self, tmpdir, data_test_dir):
        image_data = LoadStackImage.load([os.path.join(data_test_dir, "test_nucleus.tif")])
        algorithm = ThresholdAlgorithm()
        algorithm.set_image(image_data.image)
        param = algorithm.get_default_values()
        param.channel = 0
        algorithm.set_parameters(param)
        res = algorithm.calculation_run(lambda x, y: None)
        num = np.max(res.roi) + 1
        data_dict = {str(i): deepcopy(res.parameters) for i in range(1, num)}

        to_save = MaskProjectTuple(
            file_path=image_data.image.file_path,
            image=image_data.image,
            mask=None,
            roi_info=res.roi_info,
            selected_components=list(range(1, num)),
            roi_extraction_parameters=data_dict,
        )

        SaveROI.save(os.path.join(tmpdir, "segmentation2.seg"), to_save, {"relative_path": False})
        seg2 = LoadROI.load([os.path.join(tmpdir, "segmentation2.seg")])
        assert seg2 is not None

    def test_load_mask(self, data_test_dir):
        res = LoadStackImage.load([os.path.join(data_test_dir, "test_nucleus.tif")])
        assert res.mask is None
        res = LoadStackImageWithMask.load(
            [os.path.join(data_test_dir, "test_nucleus.tif"), os.path.join(data_test_dir, "test_nucleus_mask.tif")]
        )
        assert res.image.mask is not None
        assert res.mask is not None

    def test_save_project_with_history(self, tmp_path, stack_segmentation1, mask_property):
        SaveROI.save(tmp_path / "test1.seg", stack_segmentation1, {"relative_path": False})
        seg2 = dataclasses.replace(
            stack_segmentation1,
            history=[create_history_element_from_segmentation_tuple(stack_segmentation1, mask_property)],
            selected_components=[1],
            mask=stack_segmentation1.roi_info.roi,
        )
        SaveROI.save(tmp_path / "test1.seg", seg2, {"relative_path": False})
        with tarfile.open(tmp_path / "test1.seg") as tf:
            tf.getmember("mask.tif")
            tf.getmember("segmentation.tif")
            tf.getmember("history/history.json")
            tf.getmember("history/arrays_0.npz")

    def test_load_project_with_history(self, tmp_path, stack_segmentation1, mask_property):
        image_location = tmp_path / "test1.tif"
        SaveAsTiff.save(image_location, stack_segmentation1)
        seg2 = dataclasses.replace(
            stack_segmentation1,
            history=[create_history_element_from_segmentation_tuple(stack_segmentation1, mask_property)],
            selected_components=[1],
            mask=stack_segmentation1.roi_info.roi,
            image=stack_segmentation1.image.substitute(file_path=image_location),
            file_path=image_location,
        )
        SaveROI.save(tmp_path / "test1.seg", seg2, {"relative_path": False})
        res = LoadROI.load([tmp_path / "test1.seg"])
        assert res.image == str(image_location)
        assert res.mask is not None
        assert len(res.history) == 1
        assert res.history[0].mask_property == mask_property
        cmp_dict = {str(k): v for k, v in stack_segmentation1.roi_extraction_parameters.items()}
        assert str(res.history[0].roi_extraction_parameters["parameters"]) == str(cmp_dict)

    def test_mask_project_tuple(self):
        mask = np.zeros((10, 10), dtype=np.uint8)
        mask[1:-1, 1:-1] = 2
        mask2 = np.copy(mask)
        mask2[2:-2, 2:-2] = 4
        roi_info = ROIInfo(mask2)
        mask_prop = MaskProperty.simple_mask()
        elem = HistoryElement.create(roi_info, mask, {}, mask_prop)
        proj = MaskProjectTuple(
            file_path="test_data.tiff",
            image=Image(np.zeros((10, 10), dtype=np.uint8), (1, 1), "", axes_order="YX"),
            mask=mask,
            roi_info=roi_info,
            history=[elem],
            selected_components=[1, 2],
            roi_extraction_parameters={},
        )
        assert not proj.is_raw()
        assert proj.is_masked()
        raw_proj = proj.get_raw_copy()
        assert raw_proj.is_raw()
        assert not raw_proj.is_masked()
        raw_masked_proj = proj.get_raw_mask_copy()
        assert raw_masked_proj.is_masked()
        assert raw_masked_proj.is_raw()


class TestSaveFunctions:
    @staticmethod
    def read_cmap(file_path):
        with h5py.File(file_path, "r") as fp:
            arr = np.array(fp.get("Chimera/image1/data_zyx"))
            steps = tuple(int(x + 0.5) for x in fp.get("Chimera/image1").attrs["step"])
            return arr, steps

    def test_save_cmap(self, tmpdir, analysis_project):
        parameters = {"channel": 0, "separated_objects": False, "clip": False, "units": Units.nm, "reverse": False}
        SaveCmap.save(os.path.join(tmpdir, "test1.cmap"), analysis_project, parameters)
        arr, steps = self.read_cmap(os.path.join(tmpdir, "test1.cmap"))
        points = np.nonzero(arr)
        assert tuple(np.min(points, axis=1)) == (15, 15, 15)
        assert tuple(np.max(points, axis=1)) == (34, 84, 84)
        assert (1, *arr.shape) == analysis_project.roi_info.roi.shape
        assert steps == (5, 5, 10)

        parameters = {"channel": 0, "separated_objects": True, "clip": False, "units": Units.nm, "reverse": False}
        SaveCmap.save(os.path.join(tmpdir, "test2.cmap"), analysis_project, parameters)
        arr, steps = self.read_cmap(os.path.join(tmpdir, "test2_comp1.cmap"))
        points = np.nonzero(arr)
        assert tuple(np.min(points, axis=1)) == (15, 15, 15)
        assert tuple(np.max(points, axis=1)) == (34, 34, 84)
        assert (1, *arr.shape) == analysis_project.roi_info.roi.shape
        assert steps == (5, 5, 10)
        arr, steps = self.read_cmap(os.path.join(tmpdir, "test2_comp2.cmap"))
        points = np.nonzero(arr)
        assert tuple(np.min(points, axis=1)) == (15, 55, 15)
        assert tuple(np.max(points, axis=1)) == (34, 84, 84)
        assert (1, *arr.shape) == analysis_project.roi_info.roi.shape
        assert steps == (5, 5, 10)

        parameters = {"channel": 0, "separated_objects": False, "clip": True, "units": Units.nm, "reverse": False}
        SaveCmap.save(os.path.join(tmpdir, "test3.cmap"), analysis_project, parameters)
        arr, steps = self.read_cmap(os.path.join(tmpdir, "test3.cmap"))
        points = np.nonzero(arr)
        assert tuple(np.min(points, axis=1)) == (0, 0, 0)
        assert tuple(np.max(points, axis=1)) == (19, 69, 69)
        assert arr.shape == (20, 70, 70)
        assert steps == (5, 5, 10)

        parameters = {"channel": 0, "separated_objects": True, "clip": True, "units": Units.nm, "reverse": False}
        SaveCmap.save(os.path.join(tmpdir, "test4.cmap"), analysis_project, parameters)
        arr, steps = self.read_cmap(os.path.join(tmpdir, "test4_comp1.cmap"))
        points = np.nonzero(arr)
        assert tuple(np.min(points, axis=1)) == (0, 0, 0)
        assert tuple(np.max(points, axis=1)) == (19, 19, 69)
        assert arr.shape == (20, 70, 70)
        assert steps == (5, 5, 10)
        arr, steps = self.read_cmap(os.path.join(tmpdir, "test4_comp2.cmap"))
        points = np.nonzero(arr)
        assert tuple(np.min(points, axis=1)) == (0, 40, 0)
        assert tuple(np.max(points, axis=1)) == (19, 69, 69)
        assert arr.shape == (20, 70, 70)
        assert steps == (5, 5, 10)

    def test_save_cmap_reversed(self, tmpdir, analysis_project_reversed):
        parameters = {"channel": 0, "separated_objects": False, "clip": False, "units": Units.nm, "reverse": True}
        SaveCmap.save(os.path.join(tmpdir, "test1.cmap"), analysis_project_reversed, parameters)
        arr, steps = self.read_cmap(os.path.join(tmpdir, "test1.cmap"))
        points = np.nonzero(arr)
        assert tuple(np.min(points, axis=1)) == (15, 15, 15)
        assert tuple(np.max(points, axis=1)) == (34, 84, 84)
        assert (1, *arr.shape) == analysis_project_reversed.roi_info.roi.shape
        assert steps == (5, 5, 10)

        parameters = {"channel": 0, "separated_objects": True, "clip": True, "units": Units.nm, "reverse": True}
        SaveCmap.save(os.path.join(tmpdir, "test2.cmap"), analysis_project_reversed, parameters)
        arr, steps = self.read_cmap(os.path.join(tmpdir, "test2_comp1.cmap"))
        points = np.nonzero(arr)
        assert tuple(np.min(points, axis=1)) == (0, 0, 0)
        assert tuple(np.max(points, axis=1)) == (19, 19, 69)
        assert arr.shape == (20, 70, 70)
        assert steps == (5, 5, 10)
        arr, steps = self.read_cmap(os.path.join(tmpdir, "test2_comp2.cmap"))
        points = np.nonzero(arr)
        assert tuple(np.min(points, axis=1)) == (0, 40, 0)
        assert tuple(np.max(points, axis=1)) == (19, 69, 69)
        assert arr.shape == (20, 70, 70)
        assert steps == (5, 5, 10)

    @pytest.mark.parametrize("separated_objects", [True, False])
    @pytest.mark.parametrize("clip", [True, False])
    def test_save_xyz(self, tmpdir, analysis_project, separated_objects, clip):
        parameters = {"channel": 0, "separated_objects": separated_objects, "clip": clip}
        SaveXYZ.save(os.path.join(tmpdir, "test1.xyz"), analysis_project, parameters)
        array = pd.read_csv(os.path.join(tmpdir, "test1.xyz"), dtype=np.uint16, sep=" ")
        shift = (15, 15, 15, 0) if clip else (0, 0, 0, 0)

        assert np.all(np.min(array, axis=0) == np.subtract((15, 15, 15, 60), shift))
        assert np.all(np.max(array, axis=0) == np.subtract((84, 84, 34, 70), shift))
        if separated_objects:
            array = pd.read_csv(os.path.join(tmpdir, "test1_part1.xyz"), dtype=np.uint16, sep=" ")
            assert np.all(np.min(array, axis=0) == np.subtract((15, 15, 15, 70), shift))
            assert np.all(np.max(array, axis=0) == np.subtract((84, 34, 34, 70), shift))
            array = pd.read_csv(os.path.join(tmpdir, "test1_part2.xyz"), dtype=np.uint16, sep=" ")
            assert np.all(np.min(array, axis=0) == np.subtract((15, 55, 15, 60), shift))
            assert np.all(np.max(array, axis=0) == np.subtract((84, 84, 34, 60), shift))

    def test_load_old_project(self, data_test_dir):
        load_data = LoadProject.load([os.path.join(data_test_dir, "stack1_component1.tgz")])
        assert np.max(load_data.roi_info.roi) == 2
        # TODO add more checks

    def test_load_project(self, data_test_dir):
        load_data = LoadProject.load([os.path.join(data_test_dir, "stack1_component1_1.tgz")])
        assert np.max(load_data.roi_info.roi) == 2
        # TODO add more checks

    def test_save_project(self, tmpdir, analysis_project):
        SaveProject.save(os.path.join(tmpdir, "test1.tgz"), analysis_project)
        assert os.path.exists(os.path.join(tmpdir, "test1.tgz"))
        LoadProject.load([os.path.join(tmpdir, "test1.tgz")])
        # TODO add more

    def test_save_tiff(self, tmpdir, analysis_project):
        SaveAsTiff.save(os.path.join(tmpdir, "test1.tiff"), analysis_project)
        array = tifffile.imread(os.path.join(tmpdir, "test1.tiff"))
        assert analysis_project.roi_info.roi.shape == (1, *array.shape)

    def test_save_numpy(self, tmpdir, analysis_project):
        parameters = {"squeeze": False}
        SaveAsNumpy.save(os.path.join(tmpdir, "test1.npy"), analysis_project, parameters)
        array = np.load(os.path.join(tmpdir, "test1.npy"))
        assert array.shape == (1, *analysis_project.image.shape)
        assert np.all(array == analysis_project.image.get_data())
        parameters = {"squeeze": True}
        SaveAsNumpy.save(os.path.join(tmpdir, "test2.npy"), analysis_project, parameters)
        array = np.load(os.path.join(tmpdir, "test2.npy"))
        assert (1, *array.shape) == analysis_project.roi_info.roi.shape
        assert np.all(array == analysis_project.image.get_data().squeeze())

    def test_save_segmentation_numpy(self, tmpdir, analysis_project):
        SaveROIAsNumpy.save(os.path.join(tmpdir, "test1.npy"), analysis_project)
        array = np.load(os.path.join(tmpdir, "test1.npy"))
        assert np.all(array == analysis_project.roi_info.roi)

    @pytest.mark.parametrize(
        ("klass", "ext_li"),
        [
            (SaveAsNumpy, [".npy"]),
            (SaveAsTiff, [".tiff", ".tif"]),
            (SaveCmap, [".cmap"]),
            (SaveXYZ, [".xyz", ".txt"]),
            (SaveProject, [".tgz", ".tbz2", ".gz", ".bz2"]),
            (SaveROIAsNumpy, [".npy"]),
        ],
    )
    def test_get_extensions(self, klass, ext_li):
        assert klass.get_extensions() == ext_li


def test_json_parameters_mask(stack_segmentation1, tmp_path):
    SaveParametersJSON.save(tmp_path / "test.json", stack_segmentation1)
    load_param = LoadROIParameters.load([tmp_path / "test.json"])
    assert len(load_param.roi_extraction_parameters) == 4


def test_json_parameters_mask_2(stack_segmentation1, tmp_path):
    SaveParametersJSON.save(tmp_path / "test.json", stack_segmentation1.roi_extraction_parameters[1])
    load_param = LoadROIParameters.load([tmp_path / "test.json"])
    assert len(load_param.roi_extraction_parameters) == 1


@pytest.mark.parametrize("file_path", (Path(__file__).parent.parent / "test_data" / "notebook").glob("*.json"))
def test_load_notebook_json(file_path):
    """Check if all notebook files can be loaded"""
    load_metadata_base(file_path)


@pytest.mark.parametrize(
    "file_path", list((Path(__file__).parent.parent / "test_data" / "old_saves").glob(os.path.join("*", "*", "*.json")))
)
def test_old_saves_load(file_path):
    data: ProfileDict = load_metadata_base(file_path)
    assert data.verify_data(), data.pop_errors()


def test_load_plan_form_excel(bundle_test_dir):
    data, err = LoadPlanExcel.load([bundle_test_dir / "sample_batch_output.xlsx"])
    assert err == []
    assert len(data) == 3
    assert isinstance(data["test3"], CalculationPlan)
    assert isinstance(data["test4"], CalculationPlan)
    assert isinstance(data["test3 (1)"], CalculationPlan)
    assert LoadPlanExcel.get_name_with_suffix().endswith("(*.xlsx)")
    assert LoadPlanExcel.get_short_name() == "plan_excel"


def test_load_plan_form_excel_problem(bundle_test_dir):
    data, err = LoadPlanExcel.load([bundle_test_dir / "problematic_excel_batch.xlsx"])
    assert len(err) == 1
    assert len(data) == 0
    assert "not found in register" in err[0]


def test_load_json_plan(bundle_test_dir):
    data, err = LoadPlanJson.load([bundle_test_dir / "measurements_profile.json"])
    assert err == []
    assert len(data) == 1
    assert LoadPlanJson.get_name_with_suffix().endswith("(*.json)")
    assert LoadPlanJson.get_short_name() == "plan_json"


def test_load_json_plan_problem(bundle_test_dir):
    data, err = LoadPlanJson.load([bundle_test_dir / "problematic_json_batch.json"])
    assert len(err) == 1
    assert len(data) == 0
    assert "not found in register" in err[0]


def test_load_metadata_problem():
    data, err = load_metadata_part(
        '{"__class__": "PartSegCore.universal_const.Units1", "__class_version_dkt__": {"PartSegCore.universal_const.Units": "0.0.0"}, "__values__": {"value": 2}}'  # noqa: E501
    )
    assert len(err) == 1
    assert len(data) == 0


def test_find_problematic_leafs_base():
    assert find_problematic_leafs(1) == []
    assert find_problematic_leafs({"aaa": 1, "bbb": 2}) == []
    data = {"aaa": 1, "bbb": 2, "__error__": True}
    assert find_problematic_leafs(data) == [data]


def test_find_problematic_leaf_nested():
    data = {"aaa": 1, "bbb": 2, "__error__": True, "ccc": {"ddd": 1, "eee": 2, "__error__": True}}
    assert find_problematic_leafs(data) == [data["ccc"]]


def test_find_problematic_leaf_nested_class():
    data = {
        "__class__": "CalculationPlan",
        "aaa": 1,
        "bbb": 2,
        "__error__": True,
        "__values__": {"ddd": 1, "eee": 2, "Zzzz": {"__error__": True, "aa": 1}},
    }
    assert find_problematic_leafs(data) == [data["__values__"]["Zzzz"]]


def test_find_problematic_entries_base():
    assert find_problematic_entries(1) == []
    assert find_problematic_entries({"aaa": 1, "bbb": 2}) == []
    data = {"aaa": 1, "bbb": 2, "__error__": True}
    assert find_problematic_entries(data) == [data]


def test_find_problematic_entries_nested():
    data = {"aaa": 1, "bbb": 2, "__error__": True, "ccc": {"ddd": 1, "eee": 2, "__error__": True}}
    assert find_problematic_entries(data) == [data]
    data = {"aaa": 1, "bbb": 2, "ccc": {"ddd": 1, "eee": 2, "__error__": True}}
    assert find_problematic_entries(data) == [data["ccc"]]
    data = {"aaa": 1, "bbb": 2, "ccc": {"ddd": 1, "eee": 2, "__error__": True}, "kkk": {"__error__": True, "a": 1}}
    assert find_problematic_entries(data) == [data["ccc"], data["kkk"]]


def test_load_range_algorithm(bundle_test_dir):
    data, err = LoadPlanJson.load([bundle_test_dir / "range_profile.json"])
    assert err == []
    assert len(data) == 1
    assert isinstance(data["test_range"].values.threshold, RangeThresholdSelection)
    assert data["test_range"].values.threshold.name == "Range"


def test_load_image_for_batch(data_test_dir):
    assert LoadImageForBatch.get_name()
    assert LoadImageForBatch.get_short_name()
    assert LoadImageForBatch.get_extensions()
    with pytest.raises(ValueError, match="Cannot load file"):
        # assert to not try load points
        LoadImageForBatch.load(["data.csv"])

    proj = LoadImageForBatch.load([data_test_dir / "stack1_component1.tgz"])
    assert proj.image.mask is None
    assert proj.mask is None


def test_save_base_extension_parse_no_ext():
    class Save(SaveBase):
        @classmethod
        def get_name(cls) -> str:
            return "Sample save"

    with pytest.raises(ValueError, match="No extensions"):
        Save.get_extensions()


def test_save_base_extension_parse_malformatted_ext():
    class Save(SaveBase):
        @classmethod
        def get_name(cls) -> str:
            return "Sample save (a.txt)"

    with pytest.raises(ValueError, match="Error with parsing"):
        Save.get_extensions()


def test_load_points(tmp_path):
    data_path = tmp_path / "sample.csv"
    with data_path.open("w") as fp:
        fp.write(POINTS_DATA)

    res = LoadPoints.load([data_path])
    assert res.file_path == data_path
    assert res.points.shape == (5, 4)

    assert LoadPoints.get_short_name() == "point_csv"


def test_open_tar_file_with_tarfile_object(tmp_path):
    tar_file_path = tmp_path / "test.tar"
    tar_file = tarfile.TarFile.open(tar_file_path, "w")
    tar_file.close()
    result_tar_file, result_file_path = open_tar_file(tar_file)
    assert isinstance(result_tar_file, tarfile.TarFile)
    assert result_file_path == ""


def test_open_tar_file_with_path(tmp_path):
    tar_file_path = tmp_path / "test.tar"
    tar_file = tarfile.TarFile.open(tar_file_path, "w")
    tar_file.close()
    result_tar_file, result_file_path = open_tar_file(tar_file_path)
    assert isinstance(result_tar_file, tarfile.TarFile)
    assert result_file_path == str(tar_file_path)


def test_open_tar_file_with_buffer(tmp_path):
    buffer = BytesIO()
    tar_file = tarfile.TarFile.open(fileobj=buffer, mode="w")
    tar_file.close()
    buffer.seek(0)
    result_tar_file, result_file_path = open_tar_file(buffer)
    assert isinstance(result_tar_file, tarfile.TarFile)
    assert result_file_path == ""


def test_open_tar_file_with_invalid_type(tmp_path):
    with pytest.raises(ValueError, match="wrong type of file_data argument"):
        open_tar_file(123)


def test_save_mask_as_tiff(tmp_path, analysis_segmentation2):
    file_path = tmp_path / "test.tiff"
    file_path2 = tmp_path / "test2.tiff"
    assert analysis_segmentation2.image.mask is None
    SaveMaskAsTiff.save(file_path, analysis_segmentation2)
    assert file_path.exists()
    assert tifffile.TiffFile(str(file_path)).asarray().shape == analysis_segmentation2.mask.squeeze().shape
    seg = dataclasses.replace(
        analysis_segmentation2, image=analysis_segmentation2.image.substitute(mask=analysis_segmentation2.mask)
    )
    SaveMaskAsTiff.save(file_path2, seg)
    assert file_path2.exists()
    assert tifffile.TiffFile(str(file_path2)).asarray().shape == analysis_segmentation2.mask.squeeze().shape


class TestSaveScreenshot:
    def test_get_name(self):
        assert SaveScreenshot.get_default_extension() == ".png"
        assert SaveScreenshot.get_name().startswith("Screenshot")
        assert SaveScreenshot.get_short_name() == "screenshot"
        assert SaveScreenshot.get_fields() == []
        assert {".png", ".jpg", ".jpeg"}.issubset(set(SaveScreenshot.get_extensions()))

    def test_saving(self, tmp_path):
        file_path = tmp_path / "test.png"
        data = np.zeros((100, 100, 3), dtype=np.uint8)
        SaveScreenshot.save(file_path, data)
        assert file_path.exists()
        assert file_path.stat().st_size > 0


class TestLoadROIFromTIFF:
    def test_get_name(self):
        assert LoadROIFromTIFF.get_name().startswith("ROI from tiff")
        assert LoadROIFromTIFF.get_short_name() == "roi_tiff"
        assert set(LoadROIFromTIFF.get_extensions()) == {".tiff", ".tif"}

    def test_load(self, tmp_path):
        data = np.zeros((100, 100), dtype=np.uint8)
        data[10:20, 10:20] = 1
        data[30:40, 30:40] = 2
        tifffile.imwrite(tmp_path / "test.tiff", data=data)
        res = LoadROIFromTIFF.load([tmp_path / "test.tiff"])
        assert isinstance(res, MaskProjectTuple)
        assert res.image is None
        assert res.roi_info.roi.shape == (1, 1, 100, 100)


UPDATE_NAME_JSON = """
{"problematic set": {
      "__MeasurementProfile__": true,
      "name": "problematic set",
      "chosen_fields": [
        {
          "__Serializable__": true,
          "__subtype__": "PartSegCore.analysis.measurement_base.MeasurementEntry",
          "name": "Segmentation Pixel Brightness Sum",
          "calculation_tree": {
            "__Serializable__": true,
            "__subtype__": "PartSegCore.analysis.measurement_base.Leaf",
            "name": "Pixel Brightness Sum",
            "dict": {},
            "power": 1.0,
            "area": {
              "__Enum__": true,
              "__subtype__": "PartSegCore.analysis.measurement_base.AreaType",
              "value": 1
            },
            "per_component": {
              "__Enum__": true,
              "__subtype__": "PartSegCore.analysis.measurement_base.PerComponent",
              "value": 1
            },
            "channel": null
          }
        },
        {
          "__Serializable__": true,
          "__subtype__": "PartSegCore.analysis.measurement_base.MeasurementEntry",
          "name": "Segmentation Components Number",
          "calculation_tree": {
            "__Serializable__": true,
            "__subtype__": "PartSegCore.analysis.measurement_base.Leaf",
            "name": "Components Number",
            "dict": {},
            "power": 1.0,
            "area": {
              "__Enum__": true,
              "__subtype__": "PartSegCore.analysis.measurement_base.AreaType",
              "value": 1
            },
            "per_component": {
              "__Enum__": true,
              "__subtype__": "PartSegCore.analysis.measurement_base.PerComponent",
              "value": 1
            },
            "channel": null
          }
        },
        {
          "__Serializable__": true,
          "__subtype__": "PartSegCore.analysis.measurement_base.MeasurementEntry",
          "name": "Segmentation Diameter",
          "calculation_tree": {
            "__Serializable__": true,
            "__subtype__": "PartSegCore.analysis.measurement_base.Leaf",
            "name": "Diameter",
            "dict": {},
            "power": 1.0,
            "area": {
              "__Enum__": true,
              "__subtype__": "PartSegCore.analysis.measurement_base.AreaType",
              "value": 1
            },
            "per_component": {
              "__Enum__": true,
              "__subtype__": "PartSegCore.analysis.measurement_base.PerComponent",
              "value": 1
            },
            "channel": null
          }
        }
      ],
      "name_prefix": ""
    }
  }
"""


POINTS_DATA = """index,axis-0,axis-1,axis-2,axis-3
0.0,0.0,22.0,227.90873370570543,65.07832834070409
1.0,0.0,22.0,91.94021739981048,276.7482348060973
2.0,0.0,22.0,194.83531082048773,380.3782931797794
3.0,0.0,22.0,391.07095327277943,268.6636259303818
4.0,0.0,22.0,152.20734354620717,256.1692217292996
"""



# pylint: disable=no-self-use
import numpy as np
import pytest

from PartSeg._roi_analysis.image_view import CompareImageView, ResultImageView, SynchronizeView
from PartSeg.common_gui.channel_control import ChannelProperty
from PartSegCore.roi_info import ROIInfo


@pytest.mark.windows_ci_skip()
def test_synchronize(part_settings, image2, qtbot):
    prop = ChannelProperty(part_settings, "test1")
    view1 = ResultImageView(part_settings, prop, "test1")
    view2 = CompareImageView(part_settings, prop, "test2")
    sync = SynchronizeView(view1, view2)
    qtbot.add_widget(prop)
    qtbot.add_widget(view1)
    qtbot.add_widget(view2)
    view1.show()
    view2.show()
    part_settings.image = image2
    point1 = view1.viewer.dims.point
    point2 = view2.viewer.dims.point
    sync.set_synchronize(False)
    view1.viewer.dims.set_point(len(point1) - 1, point1[-1] + 1 * view1.viewer.dims.range[-1][-1])
    assert view1.viewer.dims.point != point1
    assert view2.viewer.dims.point == point2
    sync.set_synchronize(True)
    view1.viewer.dims.set_point(len(point1) - 1, point1[-1] + 2 * view1.viewer.dims.range[-1][-1])
    assert view2.viewer.dims.point != point2
    assert view2.viewer.dims.point == view1.viewer.dims.point
    view2.viewer.dims.set_point(len(point1) - 1, point1[-1])
    assert view2.viewer.dims.point == point1
    assert view1.viewer.dims.point == point1
    view1.hide()
    view2.hide()


@pytest.mark.windows_ci_skip()
def test_synchronize_change_image_dim(part_settings, image2, image2d, qtbot):
    part_settings.image = image2d
    prop = ChannelProperty(part_settings, "test1")
    view1 = ResultImageView(part_settings, prop, "test1")
    view2 = CompareImageView(part_settings, prop, "test2")
    sync = SynchronizeView(view1, view2)
    qtbot.add_widget(prop)
    qtbot.add_widget(view1)
    qtbot.add_widget(view2)
    view1.show()
    view2.show()
    sync.set_synchronize(True)
    with qtbot.waitSignal(view1.view_changed):
        part_settings.image = image2
    view1.hide()
    view2.hide()
    qtbot.wait(50)


def test_compare_view(part_settings, image2, qtbot):
    prop = ChannelProperty(part_settings, "test")
    view = CompareImageView(part_settings, prop, "test")
    qtbot.add_widget(prop)
    qtbot.add_widget(view)
    part_settings.image = image2

    roi = np.ones(image2.get_channel(0).shape, dtype=np.uint8)

    assert view.image_info[str(image2.file_path)].roi is None
    part_settings.roi = roi
    assert view.image_info[str(image2.file_path)].roi is None
    part_settings.set_segmentation_to_compare(ROIInfo(roi))
    assert view.image_info[str(image2.file_path)].roi is not None


class TestResultImageView:
    def test_setting_properties(self, part_settings, image2, qtbot):
        prop = ChannelProperty(part_settings, "test")
        view = ResultImageView(part_settings, prop, "test")
        qtbot.add_widget(prop)
        qtbot.add_widget(view)
        part_settings.image = image2

        roi = np.ones(image2.get_channel(0).shape, dtype=np.uint8)

        assert view.image_info[str(image2.file_path)].roi is None
        with qtbot.waitSignal(part_settings.roi_changed):
            part_settings.roi = roi
        assert view.any_roi()
        assert view.image_info[str(image2.file_path)].roi is not None
        assert part_settings.get_from_profile("test.image_state.opacity") == 1
        view.opacity.setValue(0.5)
        assert part_settings.get_from_profile("test.image_state.opacity") == 0.5
        assert view.image_info[str(image2.file_path)].roi.opacity == 0.5
        part_settings.set_in_profile("test.image_state.opacity", 1)
        assert view.opacity.value() == 1
        assert view.image_info[str(image2.file_path)].roi.opacity == 1

        assert not part_settings.get_from_profile("test.image_state.only_border")
        view.only_border.setChecked(True)
        assert part_settings.get_from_profile("test.image_state.only_border")
        part_settings.set_in_profile("test.image_state.only_border", False)
        assert not view.only_border.isChecked()

    @pytest.mark.windows_ci_skip()
    def test_resize(self, part_settings, image2, qtbot):
        prop = ChannelProperty(part_settings, "test")
        view = ResultImageView(part_settings, prop, "test")
        qtbot.add_widget(prop)
        qtbot.add_widget(view)
        part_settings.image = image2

        view.show()
        view.resize(500, 500)
        assert view.btn_layout2.count() == 2
        view.resize(1000, 500)
        assert view.btn_layout2.count() == 0
        view.resize(500, 500)
        assert view.btn_layout2.count() == 2
        view.hide()

    @pytest.mark.windows_ci_skip()
    def test_with_roi_alternatives(self, part_settings, image2, qtbot):
        prop = ChannelProperty(part_settings, "test")
        view = ResultImageView(part_settings, prop, "test")
        qtbot.add_widget(prop)
        qtbot.add_widget(view)
        part_settings.image = image2
        view.show()

        roi = np.ones(image2.get_channel(0).shape, dtype=np.uint8)
        part_settings.roi = roi

        assert not view.roi_alternative_select.isVisible()
        part_settings.roi = ROIInfo(roi, alternative={"t1": roi, "t2": roi})
        assert view.roi_alternative_select.isVisible()
        assert view.roi_alternative_select.count() == 3

        assert view.roi_alternative_select.currentText() == "ROI"
        view.roi_alternative_select.setCurrentText("t1")

        view.hide()



import functools
import os.path

from PartSegCore.mask.io_functions import LoadStackImageWithMask
from PartSegCore.napari_plugins.loader import partseg_loader


def napari_get_reader(path: str):
    return next(
        (
            functools.partial(partseg_loader, LoadStackImageWithMask)
            for extension in LoadStackImageWithMask.get_extensions()
            if path.endswith(extension) and os.path.exists(LoadStackImageWithMask.get_next_file([path]))
        ),
        None,
    )



# pylint: disable=no-self-use
from unittest.mock import Mock

from PartSeg._roi_analysis.advanced_window import MeasurementSettings, MultipleInput, Properties
from PartSeg._roi_analysis.advanced_window import QInputDialog as advanced_module_input
from PartSeg._roi_analysis.advanced_window import QMessageBox as advanced_message_box
from PartSegCore.analysis import AnalysisAlgorithmSelection


class TestProperties:
    def test_synchronize_voxel_size(self, qtbot, part_settings):
        widget = Properties(part_settings)
        qtbot.addWidget(widget)
        widget.lock_spacing.setChecked(True)
        widget.update_spacing()
        value = widget.spacing[1].value()
        with qtbot.waitSignal(widget.spacing[2].valueChanged, timeout=10**4):
            widget.spacing[2].setValue(value - 20)
        assert widget.spacing[1].value() == value - 20

    def test_pipeline_profile_show_info(
        self, qtbot, part_settings, border_rim_profile, lower_threshold_profile, sample_pipeline
    ):
        part_settings.roi_profiles[border_rim_profile.name] = border_rim_profile
        part_settings.roi_pipelines[sample_pipeline.name] = sample_pipeline
        widget = Properties(part_settings)
        widget.show()
        qtbot.addWidget(widget)
        widget.update_profile_list()
        assert widget.profile_list.count() == 1
        part_settings.roi_profiles[lower_threshold_profile.name] = lower_threshold_profile
        widget.update_profile_list()
        assert widget.profile_list.count() == 2
        assert widget.pipeline_list.count() == 1
        assert widget.info_label.toPlainText() == ""
        with qtbot.waitSignal(widget.profile_list.currentItemChanged, timeout=10**4):
            widget.profile_list.setCurrentRow(1)
        profile = part_settings.roi_profiles[widget.profile_list.item(1).text()]
        assert widget.info_label.toPlainText() == profile.pretty_print(AnalysisAlgorithmSelection)
        widget.pipeline_list.setCurrentRow(0)
        assert widget.info_label.toPlainText() == sample_pipeline.pretty_print(AnalysisAlgorithmSelection)
        widget.hide()

    def test_delete_profile(self, qtbot, part_settings, border_rim_profile, lower_threshold_profile):
        part_settings.roi_profiles[border_rim_profile.name] = border_rim_profile
        part_settings.roi_profiles[lower_threshold_profile.name] = lower_threshold_profile
        widget = Properties(part_settings)
        widget.show()
        qtbot.addWidget(widget)
        widget.update_profile_list()
        assert widget.profile_list.count() == 2
        with qtbot.waitSignal(widget.profile_list.currentItemChanged, timeout=10**4):
            widget.profile_list.setCurrentRow(0)
        assert widget.delete_btn.isEnabled()
        with qtbot.waitSignal(widget.delete_btn.clicked):
            widget.delete_btn.click()
        assert len(part_settings.roi_profiles) == 1
        assert lower_threshold_profile.name in part_settings.roi_profiles
        widget.hide()

    def test_rename_profile(
        self, qtbot, part_settings, border_rim_profile, lower_threshold_profile, sample_pipeline, monkeypatch
    ):
        part_settings.roi_profiles[border_rim_profile.name] = border_rim_profile
        part_settings.roi_pipelines[sample_pipeline.name] = sample_pipeline
        part_settings.roi_profiles[lower_threshold_profile.name] = lower_threshold_profile
        widget = Properties(part_settings)
        widget.show()
        qtbot.addWidget(widget)
        widget.update_profile_list()
        assert widget.profile_list.count() == 2
        monkeypatch.setattr(advanced_module_input, "getText", check_text(border_rim_profile.name, "rim"))
        widget.profile_list.setCurrentRow(0)
        widget.rename_profile()
        assert widget.profile_list.item(1).text() == "rim"
        assert set(part_settings.roi_profiles.keys()) == {"rim", lower_threshold_profile.name}
        monkeypatch.setattr(advanced_module_input, "getText", check_text(sample_pipeline.name, "rim"))
        widget.pipeline_list.setCurrentRow(0)
        widget.rename_profile()
        assert widget.pipeline_list.item(0).text() == "rim"
        assert set(part_settings.roi_pipelines.keys()) == {"rim"}
        monkeypatch.setattr(advanced_module_input, "getText", check_text("rim", lower_threshold_profile.name))

        called_mock = [0]

        def mock_waring(*_):
            monkeypatch.setattr(widget, "rename_profile", _empty)
            called_mock[0] = 1
            return advanced_message_box.No

        monkeypatch.setattr(advanced_message_box, "warning", mock_waring)
        widget.profile_list.setCurrentRow(1)
        widget.rename_profile()
        assert called_mock[0] == 1
        assert widget.profile_list.item(1).text() == "rim"
        assert set(part_settings.roi_profiles.keys()) == {"rim", lower_threshold_profile.name}
        widget.hide()

    def test_multiple_files_visibility(self, qtbot, part_settings):
        widget = Properties(part_settings)
        qtbot.addWidget(widget)
        assert not part_settings.get("multiple_files_widget")
        assert not widget.multiple_files_chk.isChecked()
        with qtbot.waitSignal(widget.multiple_files_chk.stateChanged):
            widget.multiple_files_chk.setChecked(True)
        assert part_settings.get("multiple_files_widget")
        part_settings.set("multiple_files_widget", False)
        assert not widget.multiple_files_chk.isChecked()


class TestMeasurementSettings:
    def test_create(self, qtbot, part_settings):
        widget = MeasurementSettings(part_settings)
        qtbot.addWidget(widget)

    def test_base_steep(self, qtbot, part_settings):
        widget = MeasurementSettings(part_settings)
        qtbot.addWidget(widget)
        widget.show()
        widget.profile_options.setCurrentRow(0)
        assert widget.profile_options.item(0).text() == "Volume"
        assert widget.profile_options.item(1).text() == "Diameter"
        assert widget.profile_options_chosen.count() == 0
        widget.choose_option()
        assert widget.profile_options_chosen.count() == 1
        widget.choose_option()
        assert widget.profile_options_chosen.count() == 1
        widget.profile_options.setCurrentRow(1)
        assert widget.profile_options_chosen.count() == 1
        widget.choose_option()
        assert widget.profile_options_chosen.count() == 2
        widget.profile_options.setCurrentRow(0)
        widget.proportion_action()
        assert widget.profile_options_chosen.count() == 2
        widget.profile_options.setCurrentRow(1)
        widget.proportion_action()
        assert widget.profile_options_chosen.count() == 3
        assert widget.profile_options_chosen.item(2).text() == "ROI Volume/ROI Diameter"

        widget.profile_options_chosen.setCurrentRow(0)
        assert widget.profile_options_chosen.item(0).text() == "ROI Volume"
        widget.remove_element()
        assert widget.profile_options_chosen.count() == 2
        assert widget.profile_options_chosen.item(0).text() == "ROI Diameter"

        assert not widget.save_butt.isEnabled()
        with qtbot.waitSignal(widget.profile_name.textChanged):
            widget.profile_name.setText("test")
        assert widget.save_butt.isEnabled()

        assert len(part_settings.measurement_profiles) == 2
        with qtbot.waitSignal(widget.save_butt.clicked):
            widget.save_butt.click()
        assert len(part_settings.measurement_profiles) == 3

        with qtbot.waitSignal(widget.profile_name.textChanged):
            widget.profile_name.setText("")
        assert not widget.save_butt.isEnabled()

        widget.reset_action()
        assert widget.profile_options_chosen.count() == 0
        widget.hide()


def test_multiple_input(qtbot, monkeypatch):
    mock = Mock()
    monkeypatch.setattr(advanced_message_box, "warning", mock)
    widget = MultipleInput(
        text="sample text",
        help_text="help",
        objects_list=[
            ("A", str),
            ("B", int, 5),
            ("C", float, 5.0),
            ("D", int),
            ("E", float),
        ],
    )
    qtbot.addWidget(widget)
    assert widget.result is None
    widget.accept_response()
    assert widget.result is None
    mock.assert_called_once()
    mock.reset_mock()
    widget.object_dict["A"][1].setText("test")
    widget.accept_response()
    mock.assert_not_called()
    assert widget.result == {"A": "test", "B": 5, "C": 5.0, "D": 0, "E": 0.0}


def check_text(expected, to_return):
    def _check(*_, text=None, **_kwargs):
        assert text == expected
        return to_return, True

    return _check


def _empty():
    """
    Empty function for monkeypatching to prevent recursive call
    in `test_rename_profile` test
    """



# pylint: disable=no-self-use

import itertools

import numpy as np
import pytest

from PartSegCore.segmentation.border_smoothing import (
    IterativeSmoothingParams,
    IterativeVoteSmoothing,
    OpeningSmoothing,
    OpeningSmoothingParams,
    VoteSmoothing,
    VoteSmoothingParams,
)
from PartSegCore.segmentation.watershed import NeighType


class TestVoteSmoothing:
    def test_cube_sides(self):
        data = np.zeros((50, 50, 50), dtype=np.uint8)
        data[2:-2, 2:-2, 2:-2] = 1
        res = VoteSmoothing.smooth(data, VoteSmoothingParams(neighbourhood_type=NeighType.sides, support_level=1))
        assert np.all(res == data)
        res = VoteSmoothing.smooth(data, VoteSmoothingParams(neighbourhood_type=NeighType.sides, support_level=3))
        assert np.all(res == data)
        res = VoteSmoothing.smooth(data, VoteSmoothingParams(neighbourhood_type=NeighType.sides, support_level=4))
        res2 = np.copy(data)
        for pos in itertools.product([2, -3], repeat=3):
            res2[pos] = 0
        assert np.all(res2 == res)
        res = VoteSmoothing.smooth(data, VoteSmoothingParams(neighbourhood_type=NeighType.sides, support_level=5))
        res2 = np.copy(data)
        for pos in itertools.permutations([2, 2, -3, -3, slice(2, -2)], 3):
            res2[pos] = 0
        assert np.all(res2 == res)
        res = VoteSmoothing.smooth(data, VoteSmoothingParams(neighbourhood_type=NeighType.sides, support_level=6))
        res2 = np.zeros(data.shape, dtype=data.dtype)
        res2[3:-3, 3:-3, 3:-3] = 1
        assert np.all(res2 == res)

    def test_cube_edges(self):
        data = np.zeros((50, 50, 50), dtype=np.uint8)
        data[2:-2, 2:-2, 2:-2] = 1
        res = VoteSmoothing.smooth(data, VoteSmoothingParams(neighbourhood_type=NeighType.edges, support_level=1))
        assert np.all(res == data)
        res = VoteSmoothing.smooth(data, VoteSmoothingParams(neighbourhood_type=NeighType.edges, support_level=6))
        assert np.all(res == data)
        res = VoteSmoothing.smooth(data, VoteSmoothingParams(neighbourhood_type=NeighType.edges, support_level=7))
        res2 = np.copy(data)
        for pos in itertools.product([2, -3], repeat=3):
            res2[pos] = 0
        assert np.all(res2 == res)
        res = VoteSmoothing.smooth(data, VoteSmoothingParams(neighbourhood_type=NeighType.edges, support_level=9))
        assert np.all(res2 == res)
        res = VoteSmoothing.smooth(data, VoteSmoothingParams(neighbourhood_type=NeighType.edges, support_level=10))
        res2 = np.copy(data)
        for pos in itertools.permutations([2, 2, -3, -3, slice(2, -2)], 3):
            res2[pos] = 0
        assert np.all(res2 == res)
        res = VoteSmoothing.smooth(data, VoteSmoothingParams(neighbourhood_type=NeighType.edges, support_level=13))
        assert np.all(res2 == res)
        res = VoteSmoothing.smooth(data, VoteSmoothingParams(neighbourhood_type=NeighType.edges, support_level=14))
        res2 = np.zeros(data.shape, dtype=data.dtype)
        res2[3:-3, 3:-3, 3:-3] = 1
        assert np.all(res2 == res)

    def test_cube_vertex(self):
        data = np.zeros((50, 50, 50), dtype=np.uint8)
        data[2:-2, 2:-2, 2:-2] = 1
        res = VoteSmoothing.smooth(data, VoteSmoothingParams(neighbourhood_type=NeighType.vertex, support_level=1))
        assert np.all(res == data)
        res = VoteSmoothing.smooth(data, VoteSmoothingParams(neighbourhood_type=NeighType.vertex, support_level=7))
        assert np.all(res == data)
        res = VoteSmoothing.smooth(data, VoteSmoothingParams(neighbourhood_type=NeighType.vertex, support_level=8))
        res2 = np.copy(data)
        for pos in itertools.product([2, -3], repeat=3):
            res2[pos] = 0
        assert np.all(res2 == res)
        res = VoteSmoothing.smooth(data, VoteSmoothingParams(neighbourhood_type=NeighType.vertex, support_level=11))
        assert np.all(res2 == res)
        res = VoteSmoothing.smooth(data, VoteSmoothingParams(neighbourhood_type=NeighType.vertex, support_level=12))
        res2 = np.copy(data)
        for pos in itertools.permutations([2, 2, -3, -3, slice(2, -2)], 3):
            res2[pos] = 0
        assert np.all(res2 == res)
        res = VoteSmoothing.smooth(data, VoteSmoothingParams(neighbourhood_type=NeighType.vertex, support_level=17))
        assert np.all(res2 == res)
        res = VoteSmoothing.smooth(data, VoteSmoothingParams(neighbourhood_type=NeighType.vertex, support_level=18))
        res2 = np.zeros(data.shape, dtype=data.dtype)
        res2[3:-3, 3:-3, 3:-3] = 1
        assert np.all(res2 == res)

    def test_square_sides(self):
        data = np.zeros((1, 50, 50), dtype=np.uint8)
        data[:, 2:-2, 2:-2] = 1
        res = VoteSmoothing.smooth(data, VoteSmoothingParams(neighbourhood_type=NeighType.sides, support_level=1))
        assert np.all(res == data)
        res = VoteSmoothing.smooth(data, VoteSmoothingParams(neighbourhood_type=NeighType.sides, support_level=2))
        assert np.all(res == data)
        res = VoteSmoothing.smooth(data, VoteSmoothingParams(neighbourhood_type=NeighType.sides, support_level=3))
        res2 = np.copy(data)
        for pos in itertools.product([2, -3], repeat=2):
            res2[(0, *pos)] = 0
        assert np.all(res2 == res)
        res = VoteSmoothing.smooth(data, VoteSmoothingParams(neighbourhood_type=NeighType.sides, support_level=4))
        res2 = np.zeros(data.shape, dtype=data.dtype)
        res2[:, 3:-3, 3:-3] = 1
        assert np.all(res == res2)

    def test_square_edges(self):
        data = np.zeros((1, 50, 50), dtype=np.uint8)
        data[:, 2:-2, 2:-2] = 1
        res = VoteSmoothing.smooth(data, VoteSmoothingParams(neighbourhood_type=NeighType.edges, support_level=1))
        assert np.all(res == data)
        res = VoteSmoothing.smooth(data, VoteSmoothingParams(neighbourhood_type=NeighType.edges, support_level=3))
        assert np.all(res == data)
        res = VoteSmoothing.smooth(data, VoteSmoothingParams(neighbourhood_type=NeighType.edges, support_level=4))
        res2 = np.copy(data)
        for pos in itertools.product([2, -3], repeat=2):
            res2[(0, *pos)] = 0
        assert np.all(res2 == res)
        res = VoteSmoothing.smooth(data, VoteSmoothingParams(neighbourhood_type=NeighType.edges, support_level=5))
        assert np.all(res2 == res)
        res = VoteSmoothing.smooth(data, VoteSmoothingParams(neighbourhood_type=NeighType.edges, support_level=6))
        res2 = np.zeros(data.shape, dtype=data.dtype)
        res2[:, 3:-3, 3:-3] = 1
        assert np.all(res == res2)

    def test_square_vertex(self):
        data = np.zeros((1, 50, 50), dtype=np.uint8)
        data[:, 2:-2, 2:-2] = 1
        res = VoteSmoothing.smooth(data, VoteSmoothingParams(neighbourhood_type=NeighType.vertex, support_level=1))
        assert np.all(res == data)
        res = VoteSmoothing.smooth(data, VoteSmoothingParams(neighbourhood_type=NeighType.vertex, support_level=3))
        assert np.all(res == data)
        res = VoteSmoothing.smooth(data, VoteSmoothingParams(neighbourhood_type=NeighType.vertex, support_level=4))
        res2 = np.copy(data)
        for pos in itertools.product([2, -3], repeat=2):
            res2[(0, *pos)] = 0
        assert np.all(res2 == res)
        res = VoteSmoothing.smooth(data, VoteSmoothingParams(neighbourhood_type=NeighType.vertex, support_level=5))
        assert np.all(res2 == res)
        res = VoteSmoothing.smooth(data, VoteSmoothingParams(neighbourhood_type=NeighType.vertex, support_level=6))
        res2 = np.zeros(data.shape, dtype=data.dtype)
        res2[:, 3:-3, 3:-3] = 1
        assert np.all(res == res2)


def calc_cord(pos, sign, shift):
    return tuple(np.array(pos) + np.array(sign) * np.array(shift))


def generate_neighbour_sides(dist, ndim):
    return filter(lambda x: np.sum(x) <= dist, itertools.product(range(dist + 1), repeat=ndim))


def generate_neighbour_edge(dist, ndim):
    def sub_filter(arr):
        arr = np.sort(arr)
        val = 0
        mul = 1
        for el in reversed(arr):
            val += el * mul
            mul *= 2
        return val <= dist

    return filter(sub_filter, itertools.product(range(dist + 1), repeat=ndim))


class TestIterativeVoteSmoothing:
    def test_cube_sides_base(self):
        data = np.zeros((50, 50, 50), dtype=np.uint8)
        data[2:-2, 2:-2, 2:-2] = 1
        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.sides, support_level=1, max_steps=1)
        )
        assert np.all(res == data)
        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.sides, support_level=3, max_steps=1)
        )
        assert np.all(res == data)
        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.sides, support_level=4, max_steps=1)
        )
        res2 = np.copy(data)
        for pos in itertools.product([2, -3], repeat=3):
            res2[pos] = 0
        assert np.all(res2 == res)
        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.sides, support_level=5, max_steps=1)
        )
        res2 = np.copy(data)
        for pos in itertools.permutations([2, 2, -3, -3, slice(2, -2)], 3):
            res2[pos] = 0
        assert np.all(res2 == res)
        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.sides, support_level=6, max_steps=1)
        )
        res2 = np.zeros(data.shape, dtype=data.dtype)
        res2[3:-3, 3:-3, 3:-3] = 1
        assert np.all(res2 == res)

    @pytest.mark.parametrize("max_steps", range(2, 8))
    def test_cube_sides_iter_4(self, max_steps: int):
        data = np.zeros((50, 50, 50), dtype=np.uint8)
        data[2:-2, 2:-2, 2:-2] = 1

        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.sides, support_level=4, max_steps=max_steps)
        )
        res2 = np.copy(data)
        for pos in itertools.product([2, -3], repeat=3):
            sign = np.sign(pos)
            for shift in generate_neighbour_sides(max_steps - 1, 3):
                res2[calc_cord(pos, sign, shift)] = 0
        assert np.all(res2 == res)

    @pytest.mark.parametrize("max_steps", range(2, 8))
    def test_cube_sides_iter_5(self, max_steps: int):
        data = np.zeros((50, 50, 50), dtype=np.uint8)
        data[2:-2, 2:-2, 2:-2] = 1

        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.sides, support_level=5, max_steps=max_steps)
        )
        res2 = np.copy(data)
        for pos in itertools.permutations([2, 2, -3, -3, slice(2, -2)], 3):
            res2[pos] = 0
        for ind in [0, 1, 2]:
            for pos in itertools.product([2, -3], repeat=2):
                sign = np.sign(pos)
                for shift in generate_neighbour_sides(max_steps - 1, 2):
                    pos2 = list(calc_cord(pos, sign, shift))
                    pos2.insert(ind, slice(2, -2))
                    res2[tuple(pos2)] = 0
        assert np.all(res2 == res)

    @pytest.mark.parametrize("max_steps", range(2, 8))
    def test_cube_sides_iter_6(self, max_steps: int):
        data = np.zeros((50, 50, 50), dtype=np.uint8)
        data[2:-2, 2:-2, 2:-2] = 1

        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.sides, support_level=6, max_steps=max_steps)
        )
        res2 = np.zeros(data.shape, dtype=data.dtype)
        shift = 2 + max_steps
        p = slice(shift, -shift)
        res2[p, p, p] = 1
        assert np.all(res2 == res)

    def test_cube_edges_base(self):
        data = np.zeros((50, 50, 50), dtype=np.uint8)
        data[2:-2, 2:-2, 2:-2] = 1
        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.edges, support_level=1, max_steps=1)
        )
        assert np.all(res == data)
        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.edges, support_level=6, max_steps=1)
        )
        assert np.all(res == data)
        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.edges, support_level=7, max_steps=1)
        )
        res2 = np.copy(data)
        for pos in itertools.product([2, -3], repeat=3):
            res2[pos] = 0
        assert np.all(res2 == res)
        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.edges, support_level=9, max_steps=1)
        )
        assert np.all(res2 == res)
        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.edges, support_level=10, max_steps=1)
        )
        res2 = np.copy(data)
        for pos in itertools.permutations([2, 2, -3, -3, slice(2, -2)], 3):
            res2[pos] = 0
        assert np.all(res2 == res)
        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.edges, support_level=13, max_steps=1)
        )
        assert np.all(res2 == res)
        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.edges, support_level=14, max_steps=1)
        )
        res2 = np.zeros(data.shape, dtype=data.dtype)
        res2[3:-3, 3:-3, 3:-3] = 1
        assert np.all(res2 == res)

    def test_cube_edge_iter(self):
        # This test can fail if IterativeVoting do not work correctly.
        data = np.zeros((50, 50, 50), dtype=np.uint8)
        data[2:-2, 2:-2, 2:-2] = 1
        res2 = np.copy(data)
        for pos in itertools.product([2, -3], repeat=3):
            res2[pos] = 0

        for i in range(2, 4):
            res = IterativeVoteSmoothing.smooth(
                data, IterativeSmoothingParams(neighbourhood_type=NeighType.edges, support_level=7, max_steps=i)
            )
            assert np.all(res2 == res), f"Fail  on step {i}"
        for support_level in [8, 9, 10, 11, 12]:
            res2 = VoteSmoothing.smooth(
                data, VoteSmoothingParams(neighbourhood_type=NeighType.edges, support_level=support_level)
            )
            for i in range(2, 8):
                res = IterativeVoteSmoothing.smooth(
                    data,
                    IterativeSmoothingParams(
                        neighbourhood_type=NeighType.edges, support_level=support_level, max_steps=i
                    ),
                )
                res2 = VoteSmoothing.smooth(
                    res2, VoteSmoothingParams(neighbourhood_type=NeighType.edges, support_level=support_level)
                )
                assert np.all(res2 == res), f"Fail  on step {i} for support level {support_level}"

    def test_cube_vertex_base(self):
        data = np.zeros((50, 50, 50), dtype=np.uint8)
        data[2:-2, 2:-2, 2:-2] = 1
        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.vertex, support_level=1, max_steps=1)
        )
        assert np.all(res == data)
        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.vertex, support_level=7, max_steps=1)
        )
        assert np.all(res == data)
        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.vertex, support_level=8, max_steps=1)
        )
        res2 = np.copy(data)
        for pos in itertools.product([2, -3], repeat=3):
            res2[pos] = 0
        assert np.all(res2 == res)
        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.vertex, support_level=11, max_steps=1)
        )
        assert np.all(res2 == res)
        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.vertex, support_level=12, max_steps=1)
        )
        res2 = np.copy(data)
        for pos in itertools.permutations([2, 2, -3, -3, slice(2, -2)], 3):
            res2[pos] = 0
        assert np.all(res2 == res)
        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.vertex, support_level=17, max_steps=1)
        )
        assert np.all(res2 == res)
        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.vertex, support_level=18, max_steps=1)
        )
        res2 = np.zeros(data.shape, dtype=data.dtype)
        res2[3:-3, 3:-3, 3:-3] = 1
        assert np.all(res2 == res)

    def test_cube_vertex_iter(self):
        # This test can fail if IterativeVoting do not work correctly.
        data = np.zeros((50, 50, 50), dtype=np.uint8)
        data[2:-2, 2:-2, 2:-2] = 1

        for i in range(2, 4):
            res = IterativeVoteSmoothing.smooth(
                data, IterativeSmoothingParams(neighbourhood_type=NeighType.vertex, support_level=7, max_steps=i)
            )
            assert np.all(data == res), f"Fail  on step {i}"
        for support_level in range(8, 19):
            res2 = VoteSmoothing.smooth(
                data, VoteSmoothingParams(neighbourhood_type=NeighType.vertex, support_level=support_level)
            )
            for i in range(2, 8):
                res = IterativeVoteSmoothing.smooth(
                    data,
                    IterativeSmoothingParams(
                        neighbourhood_type=NeighType.vertex, support_level=support_level, max_steps=i
                    ),
                )
                res2 = VoteSmoothing.smooth(
                    res2, VoteSmoothingParams(neighbourhood_type=NeighType.vertex, support_level=support_level)
                )
                assert np.all(res2 == res), f"Fail  on step {i} for support level {support_level}"

    def test_square_sides_base(self):
        data = np.zeros((1, 50, 50), dtype=np.uint8)
        data[:, 2:-2, 2:-2] = 1
        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.sides, support_level=1, max_steps=1)
        )
        assert np.all(res == data)
        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.sides, support_level=2, max_steps=1)
        )
        assert np.all(res == data)
        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.sides, support_level=3, max_steps=1)
        )
        res2 = np.copy(data)
        for pos in itertools.product([2, -3], repeat=2):
            res2[(0, *pos)] = 0
        assert np.all(res2 == res)
        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.sides, support_level=4, max_steps=1)
        )
        res2 = np.zeros(data.shape, dtype=data.dtype)
        res2[:, 3:-3, 3:-3] = 1
        assert np.all(res == res2)

    def test_square_sides_iterative(self):
        data = np.zeros((1, 50, 50), dtype=np.uint8)
        data[:, 2:-2, 2:-2] = 1
        for i in range(2, 4):
            res = IterativeVoteSmoothing.smooth(
                data, IterativeSmoothingParams(neighbourhood_type=NeighType.sides, support_level=3, max_steps=i)
            )
            res2 = np.copy(data)
            for pos in itertools.product([2, -3], repeat=2):
                sign = np.sign(pos)
                for shift in generate_neighbour_sides(i - 1, 2):
                    res2[(0, *calc_cord(pos, sign, shift))] = 0
            assert np.all(res == res2)

        for i in range(2, 8):
            res = IterativeVoteSmoothing.smooth(
                data, IterativeSmoothingParams(neighbourhood_type=NeighType.sides, support_level=4, max_steps=i)
            )
            res2 = np.zeros(data.shape, dtype=data.dtype)
            shift = 2 + i
            p = slice(shift, -shift)
            res2[:, p, p] = 1
            assert np.all(res2 == res)

    def test_square_edges_base(self):
        data = np.zeros((1, 50, 50), dtype=np.uint8)
        data[:, 2:-2, 2:-2] = 1
        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.edges, support_level=1, max_steps=1)
        )
        assert np.all(res == data)
        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.edges, support_level=3, max_steps=1)
        )
        assert np.all(res == data)
        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.edges, support_level=4, max_steps=1)
        )
        res2 = np.copy(data)
        for pos in itertools.product([2, -3], repeat=2):
            res2[(0, *pos)] = 0
        assert np.all(res2 == res)
        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.edges, support_level=5, max_steps=1)
        )
        assert np.all(res2 == res)
        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.edges, support_level=6, max_steps=1)
        )
        res2 = np.zeros(data.shape, dtype=data.dtype)
        res2[:, 3:-3, 3:-3] = 1
        assert np.all(res == res2)

    def test_square_edges_iterative(self):
        data = np.zeros((1, 50, 50), dtype=np.uint8)
        data[:, 2:-2, 2:-2] = 1

        res2 = np.copy(data)
        for pos in itertools.product([2, -3], repeat=2):
            res2[(0, *pos)] = 0
        for _ in range(2, 5):
            res = IterativeVoteSmoothing.smooth(
                data, IterativeSmoothingParams(neighbourhood_type=NeighType.edges, support_level=4, max_steps=1)
            )
            assert np.all(res2 == res)
        for i in range(2, 8):
            res = IterativeVoteSmoothing.smooth(
                data, IterativeSmoothingParams(neighbourhood_type=NeighType.edges, support_level=5, max_steps=i)
            )
            res2 = np.copy(data)
            for pos in itertools.product([2, -3], repeat=2):
                sign = np.sign(pos)
                for shift in generate_neighbour_edge(i - 1, 2):
                    res2[(0, *calc_cord(pos, sign, shift))] = 0
            assert np.all(res == res2)
        for i in range(2, 8):
            res = IterativeVoteSmoothing.smooth(
                data, IterativeSmoothingParams(neighbourhood_type=NeighType.edges, support_level=6, max_steps=i)
            )
            res2 = np.zeros(data.shape, dtype=data.dtype)
            shift = 2 + i
            p = slice(shift, -shift)
            res2[:, p, p] = 1
            assert np.all(res2 == res)

    def test_square_vertex_base(self):
        data = np.zeros((1, 50, 50), dtype=np.uint8)
        data[:, 2:-2, 2:-2] = 1
        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.vertex, support_level=1, max_steps=1)
        )
        assert np.all(res == data)
        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.vertex, support_level=3, max_steps=1)
        )
        assert np.all(res == data)
        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.vertex, support_level=4, max_steps=1)
        )
        res2 = np.copy(data)
        for pos in itertools.product([2, -3], repeat=2):
            res2[(0, *pos)] = 0
        assert np.all(res2 == res)
        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.vertex, support_level=5, max_steps=1)
        )
        assert np.all(res2 == res)
        res = IterativeVoteSmoothing.smooth(
            data, IterativeSmoothingParams(neighbourhood_type=NeighType.vertex, support_level=6, max_steps=1)
        )
        res2 = np.zeros(data.shape, dtype=data.dtype)
        res2[:, 3:-3, 3:-3] = 1
        assert np.all(res == res2)

    def test_square_vertex_iterative(self):
        data = np.zeros((1, 50, 50), dtype=np.uint8)
        data[:, 2:-2, 2:-2] = 1

        res2 = np.copy(data)
        for pos in itertools.product([2, -3], repeat=2):
            res2[(0, *pos)] = 0
        for _ in range(2, 5):
            res = IterativeVoteSmoothing.smooth(
                data, IterativeSmoothingParams(neighbourhood_type=NeighType.vertex, support_level=4, max_steps=1)
            )
            assert np.all(res2 == res)
        for i in range(2, 8):
            res = IterativeVoteSmoothing.smooth(
                data, IterativeSmoothingParams(neighbourhood_type=NeighType.vertex, support_level=5, max_steps=i)
            )
            res2 = np.copy(data)
            for pos in itertools.product([2, -3], repeat=2):
                sign = np.sign(pos)
                for shift in generate_neighbour_edge(i - 1, 2):
                    res2[(0, *calc_cord(pos, sign, shift))] = 0
            assert np.all(res == res2)
        for i in range(2, 8):
            res = IterativeVoteSmoothing.smooth(
                data, IterativeSmoothingParams(neighbourhood_type=NeighType.vertex, support_level=6, max_steps=i)
            )
            res2 = np.zeros(data.shape, dtype=data.dtype)
            shift = 2 + i
            p = slice(shift, -shift)
            res2[:, p, p] = 1
            assert np.all(res2 == res)


class TestOpeningSmooth:
    def test_cube(self):
        data = np.zeros((50, 50, 50), dtype=np.uint8)
        data[2:-2, 2:-2, 2:-2] = 1
        res = OpeningSmoothing.smooth(data, OpeningSmoothingParams(smooth_border_radius=1))
        res2 = np.copy(data)
        for pos in itertools.product([2, -3], repeat=3):
            res2[pos] = 0
        assert np.all(res2 == res)

        res = OpeningSmoothing.smooth(data, OpeningSmoothingParams(smooth_border_radius=2))
        res2 = np.copy(data)
        for pos in itertools.permutations([2, 2, -3, -3, slice(2, -2)], 3):
            res2[pos] = 0
        assert np.all(res2 == res)

    def test_sides(self):
        data = np.zeros((1, 50, 50), dtype=np.uint8)
        data[:, 2:-2, 2:-2] = 1
        res = OpeningSmoothing.smooth(data, OpeningSmoothingParams(smooth_border_radius=1))
        assert np.all(data == res)
        res = OpeningSmoothing.smooth(data, OpeningSmoothingParams(smooth_border_radius=2))
        res2 = np.copy(data)
        for pos in itertools.product([2, -3], repeat=2):
            res2[(0, *pos)] = 0
        assert np.all(res2 == res)



"""
This module contains widgets to create and manage labels scheme
"""

import json
import typing
from copy import deepcopy
from io import BytesIO
from pathlib import Path
from typing import Any, Callable, List, Optional, Sequence, Union

import numpy as np
from fonticon_fa6 import FA6S
from qtpy.QtCore import Qt, Signal, Slot
from qtpy.QtGui import QColor, QMouseEvent, QPainter, QPaintEvent
from qtpy.QtWidgets import (
    QApplication,
    QColorDialog,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QRadioButton,
    QToolButton,
    QVBoxLayout,
    QWidget,
)
from superqt.fonticon import setTextIcon

from PartSeg.common_backend.base_settings import ViewSettings
from PartSeg.common_gui.custom_load_dialog import PLoadDialog
from PartSeg.common_gui.custom_save_dialog import PSaveDialog
from PartSeg.common_gui.icon_selector import IconSelector
from PartSeg.common_gui.numpy_qimage import NumpyQImage
from PartSegCore.custom_name_generate import custom_name_generate
from PartSegCore.io_utils import IO_LABELS_COLORMAP, LoadBase, SaveBase
from PartSegCore.utils import BaseModel

_icon_selector = IconSelector()


def add_alpha_channel(colors):
    if colors.shape[1] == 4:
        return colors
    new_label = np.zeros((colors.shape[0], 4), dtype=np.uint8)
    new_label[:, :3] = colors
    new_label[:, 3] = 255
    return new_label


class _LabelShow(QWidget):
    def __init__(self, label: np.ndarray):
        super().__init__()
        self.image = None
        self.set_labels(label)

    def set_labels(self, label):
        if label.ndim != 2 and label.shape[1] not in (3, 4):
            raise ValueError("Wrong array shape")  # pragma: no cover
        label = add_alpha_channel(label)
        self.image = NumpyQImage(label.reshape((1, *label.shape)))
        self.repaint()

    def paintEvent(self, event: QPaintEvent):
        painter = QPainter(self)
        rect = self.rect()
        painter.drawRect(rect)
        painter.drawImage(rect, self.image)


class LabelShow(QWidget):
    """Present single label color scheme"""

    remove_labels = Signal(str)
    edit_labels = Signal(list)
    edit_labels_with_name = Signal(str, list)
    selected = Signal(str)

    def __init__(self, name: str, label: List[Sequence[float]], removable, parent=None):
        super().__init__(parent)
        self.label = label
        self.name = name
        self.removable = removable

        self.radio_btn = QRadioButton()

        self.label_show = _LabelShow(np.array(label, dtype=np.uint8))

        self.remove_btn = QToolButton()
        setTextIcon(self.remove_btn, FA6S.trash_can, 16)

        self.edit_btn = QToolButton()
        setTextIcon(self.edit_btn, FA6S.pen, 16)

        if removable:
            self.remove_btn.setToolTip("Remove colormap")
        else:
            self.remove_btn.setToolTip("This colormap is protected")

        self.edit_btn.setToolTip("Create new label schema base on this")

        layout = QHBoxLayout()
        layout.addWidget(self.radio_btn)
        layout.addWidget(self.label_show, 1)
        layout.addWidget(self.remove_btn)
        layout.addWidget(self.edit_btn)
        self.remove_btn.setEnabled(removable)
        self.setLayout(layout)
        self.remove_btn.clicked.connect(self.remove_fun)
        self.edit_btn.clicked.connect(self.edit_fun)
        self.radio_btn.toggled.connect(self.selected_fun)

    def set_checked(self, val):
        self.radio_btn.setChecked(val)
        if self.removable:
            self.remove_btn.setDisabled(val)

    @Slot()
    def remove_fun(self):
        if self.remove_btn.isEnabled():
            self.remove_labels.emit(self.name)

    @Slot()
    def edit_fun(self):
        self.edit_labels_with_name.emit(self.name, deepcopy(self.label))
        self.edit_labels.emit(deepcopy(self.label))

    @Slot(bool)
    def selected_fun(self):
        if self.radio_btn.isChecked():
            self.selected.emit(self.name)


class LabelChoose(QWidget):
    edit_signal = Signal(list)
    edit_with_name_signal = Signal(str, list)

    def __init__(self, settings, parent=None):
        super().__init__(parent)
        self.settings = settings
        layout = QVBoxLayout()
        self.setLayout(layout)

    @Slot(str)
    def change_scheme(self, name):
        self.settings.current_labels = name
        for i in range(self.layout().count()):
            el = self.layout().itemAt(i)
            if el.widget():
                if el.widget().name != name:
                    el.widget().set_checked(False)
                else:
                    el.widget().set_checked(True)

    @Slot(str)
    def remove(self, name: str):
        if name in self.settings.label_color_dict:
            del self.settings.label_color_dict[name]
        self.refresh()

    def refresh(self):
        for _ in range(self.layout().count()):
            el = self.layout().takeAt(0)
            if el.widget():
                w: LabelShow = el.widget()
                w.selected.disconnect()
                w.remove_labels.disconnect()
                w.edit_labels_with_name.disconnect()
                w.edit_labels.disconnect()
                el.widget().deleteLater()

        chosen_name = self.settings.current_labels
        for name, (val, removable) in self.settings.label_color_dict.items():
            label = self._label_show(name, val, removable)
            if name == chosen_name:
                label.set_checked(True)
            label.selected.connect(self.change_scheme)
            label.remove_labels.connect(self.remove)
            label.edit_labels.connect(self.edit_signal.emit)
            label.edit_labels_with_name.connect(self.edit_with_name_signal.emit)
            self.layout().addWidget(label)
        self.layout().addStretch(1)

    def _label_show(self, name: str, label: List[Sequence[float]], removable) -> LabelShow:
        return LabelShow(name, label, removable, self)

    def showEvent(self, _):
        self.refresh()


class ColorShow(QLabel):
    """Widget to show chosen color. Change mouse cursor when above."""

    def __init__(self, color, parent=None):
        super().__init__(parent)
        self.color = color
        self._qcolor = QColor(*color)

    def paintEvent(self, event: QPaintEvent):
        painter = QPainter(self)
        painter.fillRect(event.rect(), self._qcolor)

    def enterEvent(self, _event):  # pylint: disable=no-self-use
        QApplication.setOverrideCursor(Qt.CursorShape.DragMoveCursor)

    def leaveEvent(self, _event):  # pylint: disable=no-self-use
        QApplication.restoreOverrideCursor()

    def set_color(self, color):
        self.color = color
        self._qcolor = QColor(*color)
        self.repaint()


class LabelEditor(QWidget):
    """Widget for create label scheme."""

    def __init__(self, settings: ViewSettings, parent=None):
        super().__init__(parent=parent)
        self.settings = settings
        self.color_list = []
        self.chosen = None
        self.prohibited_names = set(self.settings.label_color_dict.keys())  # Prohibited name is added to reduce
        # probability of colormap cache collision

        self.color_picker = QColorDialog()
        self.color_picker.setWindowFlag(Qt.WindowType.Widget)
        self.color_picker.setOptions(
            QColorDialog.ColorDialogOption.DontUseNativeDialog | QColorDialog.ColorDialogOption.NoButtons
        )
        self.add_color_btn = QPushButton("Add color")
        self.add_color_btn.clicked.connect(self.add_color)
        self.remove_color_btn = QPushButton("Remove last color")
        self.remove_color_btn.clicked.connect(self.remove_color)
        self.save_btn = QPushButton("Save")
        self.save_btn.clicked.connect(self.save)
        self.import_btn = QPushButton("Import")
        self.import_btn.clicked.connect(self._import_action)
        self.export_btn = QPushButton("Export")
        self.export_btn.clicked.connect(self._export_action)

        self.color_layout = QHBoxLayout()
        layout = QVBoxLayout()
        layout.addWidget(self.color_picker)
        btn_layout = QHBoxLayout()
        btn_layout.addWidget(self.add_color_btn)
        btn_layout.addWidget(self.remove_color_btn)
        btn_layout.addWidget(self.import_btn)
        btn_layout.addWidget(self.export_btn)
        btn_layout.addWidget(self.save_btn)
        layout.addLayout(btn_layout)
        layout.addLayout(self.color_layout)
        self.setLayout(layout)

    def _import_action(self):
        dial = PLoadDialog(LabelsLoad, settings=self.settings, path=IO_LABELS_COLORMAP)
        if dial.exec_():
            res = dial.get_result()
            self.set_colors("", res.load_class.load(res.load_location))

    def _export_action(self):
        if not self.color_layout.count():
            return
        self.get_colors()
        dial = PSaveDialog(
            LabelsSave,
            settings=self.settings,
            path=IO_LABELS_COLORMAP,
        )
        if dial.exec_():
            res = dial.get_result()
            res.save_class.save(res.save_destination, self.get_colors(), res.parameters)

    @Slot(str, list)
    def set_colors(self, name: str, colors: list):
        for _ in range(self.color_layout.count()):
            el = self.color_layout.takeAt(0)
            if el.widget():
                el.widget().deleteLater()
        for color in colors:
            self.color_layout.addWidget(ColorShow(color, self))

    def remove_color(self):
        if self.color_layout.count():
            el = self.color_layout.takeAt(self.color_layout.count() - 1)
            el.widget().deleteLater()

    def add_color(self):
        color = self.color_picker.currentColor()
        self.color_layout.addWidget(ColorShow([color.red(), color.green(), color.blue()], self))

    def get_colors(self) -> List[List[int]]:
        count = self.color_layout.count()
        return [self.color_layout.itemAt(i).widget().color for i in range(count)]

    def save(self):
        count = self.color_layout.count()
        if not count:
            return
        rand_name = custom_name_generate(self.prohibited_names, self.settings.label_color_dict)
        self.prohibited_names.add(rand_name)
        self.settings.label_color_dict[rand_name] = self.get_colors()

    def mousePressEvent(self, e: QMouseEvent):
        child = self.childAt(e.pos())
        if not isinstance(child, ColorShow):
            self.chosen = None
            return
        self.chosen = child

    def mouseMoveEvent(self, e: QMouseEvent):
        if self.chosen is None:
            return
        index = self.color_layout.indexOf(self.chosen)
        index2 = int(e.x() / self.width() * self.color_layout.count() + 0.5)
        if index2 != index:
            self.color_layout.insertWidget(index2, self.chosen)

    def mouseReleaseEvent(self, e: QMouseEvent):
        self.chosen = None


class LabelsLoad(LoadBase):
    __argument_class__ = BaseModel

    @classmethod
    def get_name(cls) -> str:
        return "Labels json (*.label.json)"

    @classmethod
    def load(
        cls,
        load_locations: List[Union[str, BytesIO, Path]],
        range_changed: Optional[Callable[[int, int], Any]] = None,
        step_changed: Optional[Callable[[int], Any]] = None,
        metadata: Optional[dict] = None,
    ) -> List[List[float]]:
        with open(load_locations[0]) as f_p:
            return json.load(f_p)

    @classmethod
    def get_short_name(cls):
        return "label_json"


class LabelsSave(SaveBase):
    __argument_class__ = BaseModel

    @classmethod
    def get_short_name(cls):
        return "label_json"

    @classmethod
    def save(
        cls,
        save_location: typing.Union[str, BytesIO, Path],
        project_info,
        parameters: Optional[dict] = None,
        range_changed=None,
        step_changed=None,
    ):
        with open(save_location, "w") as f_p:
            json.dump(project_info, f_p)

    @classmethod
    def get_name(cls) -> str:
        return "Labels json (*.label.json)"



import os
import typing
from contextlib import suppress
from pathlib import Path

from qtpy.QtWidgets import QDialog, QFileDialog, QGridLayout, QPushButton, QStackedWidget

from PartSeg.common_gui.algorithms_description import FormWidget
from PartSeg.common_gui.custom_load_dialog import IORegister, LoadRegisterFileDialog
from PartSegCore.algorithm_describe_base import get_fields_from_algorithm
from PartSegCore.io_utils import SaveBase

if typing.TYPE_CHECKING:  # pragma: no cover
    from PartSeg.common_backend.base_settings import BaseSettings


class SaveProperty(typing.NamedTuple):
    save_destination: typing.Union[str, typing.List[str]]
    selected_filter: str
    save_class: SaveBase
    parameters: dict


class FormDialog(QDialog):
    @staticmethod
    def widget_class() -> typing.Type[FormWidget]:
        return FormWidget

    def __init__(self, fields, values=None, image=None, settings=None, parent=None):
        super().__init__(parent)
        self.widget = self.widget_class()(fields, settings=settings)
        if values is not None:
            self.widget.set_values(values)
        if image is not None:
            self.widget.image_changed(image)
        self.accept_btn = QPushButton("Save")
        self.accept_btn.clicked.connect(self.accept)
        self.reject_btn = QPushButton("Reject")
        self.reject_btn.clicked.connect(self.reject)
        layout = QGridLayout()
        layout.addWidget(self.widget, 0, 0, 1, 2)
        layout.addWidget(self.reject_btn, 1, 0)
        layout.addWidget(self.accept_btn, 1, 1)
        self.setLayout(layout)

    def get_values(self):
        return self.widget.get_values()

    def set_values(self, values):
        return self.widget.set_values(values)


class CustomSaveDialog(LoadRegisterFileDialog):
    def __init__(
        self,
        save_register: IORegister,
        system_widget=True,
        base_values: typing.Optional[dict] = None,
        parent=None,
        caption="Save file",
        history: typing.Optional[typing.List[str]] = None,
        file_mode=QFileDialog.FileMode.AnyFile,
    ):
        super().__init__(save_register, caption, parent)
        self.setFileMode(file_mode)
        self.setOption(QFileDialog.Option.DontUseNativeDialog, not system_widget)
        self.setAcceptMode(QFileDialog.AcceptMode.AcceptSave)
        self.filterSelected.connect(self.change_filter)
        self.accepted_native = False
        self.values = {}
        self.names = []
        if history is not None:
            history = self.history() + history
            self.setHistory(history)
        self.base_values = base_values if base_values is not None else {}
        if not system_widget:
            widget = QStackedWidget()
            for name, val in self.io_register.items():
                wi = FormWidget(val)
                if name in self.base_values:
                    wi.set_values(self.base_values[name])
                widget.addWidget(wi)
                self.names.append(name)

            self.filterSelected.connect(self.change_parameters)

            layout = self.layout()
            if isinstance(layout, QGridLayout):
                # noinspection PyArgumentList
                layout.addWidget(widget, 0, layout.columnCount(), layout.rowCount(), 1)
            else:
                layout.addWidget(widget)
            self.stack_widget = widget
            self.selectNameFilter(self.names[0])

    def change_parameters(self, text):
        if not hasattr(self, "stack_widget"):
            return
        with suppress(ValueError):
            self.stack_widget.setCurrentIndex(self.names.index(text))
            if typing.cast(FormWidget, self.stack_widget.currentWidget()).has_elements():
                self.stack_widget.show()
            else:
                self.stack_widget.hide()

    def selectNameFilter(self, filter_name: str):
        with suppress(IndexError):
            self.change_parameters(filter_name)
        super().selectNameFilter(filter_name)
        with suppress(KeyError):
            ext = self.io_register[filter_name].get_default_extension()
            self.setDefaultSuffix(ext)

    def change_filter(self, current_filter):
        if current_filter not in self.io_register:
            return
        ext = self.io_register[current_filter].get_default_extension()
        self.setDefaultSuffix(ext)

    def accept(self):
        self.accepted_native = True
        if hasattr(self, "stack_widget"):
            self.values = typing.cast(FormWidget, self.stack_widget.currentWidget()).get_values()
            super().accept()
            return
        save_class = self.io_register[self.selectedNameFilter()]
        fields = get_fields_from_algorithm(save_class)
        if len(fields) == 0:
            super().accept()
            return
        dial = FormDialog(fields)
        if self.selectedNameFilter() in self.base_values:
            dial.set_values(self.base_values[self.selectedNameFilter()])
        if dial.exec_():
            self.values = dial.get_values()
            super().accept()
        else:
            super().reject()

    def get_result(self) -> SaveProperty:
        files = self.selectedFiles()
        return SaveProperty(
            files[0] if len(files) == 1 else files,
            self.selectedNameFilter(),
            self.io_register[self.selectedNameFilter()],
            self.values,
        )


class PSaveDialog(CustomSaveDialog):
    def __init__(
        self,
        save_register: IORegister,
        *,
        settings: "BaseSettings",
        path: str,
        default_directory: typing.Optional[str] = None,
        filter_path="",
        system_widget=True,
        base_values: typing.Optional[dict] = None,
        parent=None,
        caption="Save file",
        file_mode=QFileDialog.FileMode.AnyFile,
    ):
        if default_directory is None:
            default_directory = str(Path.home())
        super().__init__(
            save_register=save_register,
            system_widget=system_widget,
            base_values=base_values,
            parent=parent,
            caption=caption,
            history=settings.get_path_history(),
            file_mode=file_mode,
        )
        self.settings = settings
        self.path_in_dict = path
        self.filter_path = filter_path
        self.setDirectory(self.settings.get(path, default_directory))
        if self.filter_path:
            self.selectNameFilter(self.settings.get(self.filter_path, ""))

    def accept(self):
        super().accept()
        if self.result() != QDialog.DialogCode.Accepted:
            return
        directory = os.path.dirname(self.selectedFiles()[0])
        self.settings.add_path_history(directory)
        self.settings.set(self.path_in_dict, directory)
        if self.filter_path:
            self.settings.set(self.filter_path, self.selectedNameFilter())


SaveDialog = CustomSaveDialog



from PartSeg._roi_analysis.batch_window import (
    BatchWindow,
)


class TestBatchWindow:
    def test_create(self, part_settings):
        dlg = BatchWindow(part_settings)
        assert not dlg.is_working()



from importlib.metadata import PackageNotFoundError, version
from typing import TYPE_CHECKING, Dict, Union

try:
    PYDANTIC_2 = version("pydantic") >= "2.0.0"
except PackageNotFoundError:  # pragma: no cover
    PYDANTIC_2 = False


if TYPE_CHECKING:
    from pydantic import GetJsonSchemaHandler
    from pydantic_core.core_schema import CoreSchema


def check_type(value):  # type: ignore [misc]
    if isinstance(value, Channel):
        return value
    if value.__class__.__module__.startswith("napari"):
        value = value.name
    if not isinstance(value, (str, int)):
        raise TypeError(f"Channel need to be int or str, provided {type(value)}")
    return Channel(value)


if PYDANTIC_2:

    def check_type_(value, _validation_info=None, **_):
        return check_type(value)

else:

    def check_type_(value):  # type: ignore [misc]
        return check_type(value)


class Channel:
    """
    This class is introduced to distinguish numerical algorithm parameter from choose channel.
    In autogenerated interface field with this type limits input values to number of current image channels
    """

    def __init__(self, value: Union[str, int, "Channel"]):
        if isinstance(value, Channel):
            value = value.value
        if not isinstance(value, (str, int)):
            raise TypeError(f"wrong type {value} {type(value)}")  # pragma: no cover
        self._value: Union[str, int] = value

    @property
    def value(self) -> Union[str, int]:
        """Value stored in this class"""
        return self._value

    def __str__(self):
        return str(self._value + 1) if isinstance(self._value, int) else self._value

    def __repr__(self):
        return f"<{self.__class__.__module__}.{self.__class__.__name__}(value={self._value!r})>"

    def __eq__(self, other):
        return isinstance(other, Channel) and self._value == other.value

    def __hash__(self):
        return hash(self._value)

    def as_dict(self):
        return {"value": self._value}

    @classmethod
    def __get_validators__(cls):
        yield check_type_

    @classmethod
    def __modify_schema__(cls, field_schema):
        """Pydantic 1 dataclass schema modification method. It is used to modify schema for this class"""
        # TODO check if still required
        field_schema["title"] = "Channel"
        field_schema["type"] = "object"
        field_schema["properties"] = {"value": {"title": "value", "anyOf": [{"type": "string"}, {"type": "integer"}]}}

    @classmethod
    def __get_pydantic_json_schema__(cls, core_schema: "CoreSchema", handler: "GetJsonSchemaHandler"):
        json_schema: Dict[str, Union[str, dict]] = {}
        cls.__modify_schema__(json_schema)
        return json_schema



import typing

from qtpy.QtCore import QObject
from qtpy.QtWidgets import QWidget

from PartSeg.common_backend.abstract_class import QtMeta
from PartSeg.common_backend.partially_const_dict import PartiallyConstDict


def test_object_inheritance():
    class A:
        pass

    class B(QObject, A, metaclass=QtMeta):
        pass

    assert issubclass(B, A)
    assert issubclass(B, QObject)


def test_widget_inheritance():
    class A:
        pass

    class B(QWidget, A, metaclass=QtMeta):
        pass

    assert issubclass(B, A)
    assert issubclass(B, QWidget)


def test_object_generic_inheritance():
    T = typing.TypeVar("T")

    class A:
        pass

    class B(QWidget, A, typing.Generic[T], metaclass=QtMeta):
        pass

    assert issubclass(B, A)
    assert issubclass(B, QWidget)


def test_partial_const_dict():
    class A(PartiallyConstDict[int]):
        const_item_dict = {"a": 1, "b": 2}

    assert len(A({})) == 2
